# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for condition related operations.
"""
import unittest

import torch

import texar.torch as tx
from texar.torch.run import condition as cond
from texar.torch.run.executor_test import DummyClassifier, DummyData


class ConditionTest(unittest.TestCase):
    def _create_dataset(self, n_examples: int):
        data = torch.randint(self.vocab_size, size=(n_examples, 20))
        labels = torch.randint(self.n_classes, size=(n_examples,)).tolist()
        source = tx.data.SequenceDataSource(list(zip(data, labels)))
        dataset = DummyData(source, hparams={"batch_size": 10})
        return dataset

    def setUp(self) -> None:
        self.vocab_size = 100
        self.n_classes = 5
        self.model = DummyClassifier(self.vocab_size, self.n_classes)
        self.datasets = {
            split: self._create_dataset(n_examples)
            for split, n_examples in [
                ("train", 200), ("valid", 50), ("test1", 50), ("test2", 20)]}

    def test_once(self):
        num_iters = 2
        max_triggers = 3

        test = self

        class Recorder:
            def __init__(self):
                self.count = 0
                self.once_triggered = False

            def every_time_fn(self, executor):
                self.count += 1
                test.assertEqual(
                    executor.status["iteration"], self.count * num_iters)
                if self.count > max_triggers:
                    test.fail("Training should terminate but didn't")
                elif self.count == max_triggers:
                    executor.terminate()

            def once_fn(self, _):
                test.assertFalse(self.once_triggered)
                self.once_triggered = True

        recorder = Recorder()

        executor = tx.run.Executor(
            model=self.model,
            train_data=self.datasets["train"],
            valid_data=self.datasets["valid"],
            test_data=[self.datasets["test1"], ("t2", self.datasets["test2"])],
            optimizer={"type": torch.optim.Adam}
        )

        executor.on(cond.iteration(num_iters), recorder.every_time_fn)
        executor.on(cond.once(cond.iteration(num_iters)), recorder.once_fn)

        executor.train()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit test for gin utility functions
"""

import unittest

import os
import tempfile

from texar.torch.modules.pretrained.t5_utils import read_t5_gin_config_file


class GinTest(unittest.TestCase):
    r"""Tests for T5 gin file.

    """

    def setUp(self):
        test_gin = '''d_ff = 2048\n''' \
                   '''d_kv = 64\n''' \
                   '''d_model = 512\n'''\
                   '''dropout_rate = 0.1\n''' \
                   '''inputs_length = 512\n''' \
                   '''mean_noise_span_length = 3.0\n''' \
                   '''MIXTURE_NAME = 'all_mix'\n''' \
                   '''noise_density = 0.15\n''' \
                   '''num_heads = 8\n''' \
                   '''num_layers = 6\n''' \
                   '''targets_length = 512\n''' \
                   '''init_checkpoint = "gs://t5-data/pretrained_models/''' \
                   '''small/model.ckpt-1000000"\n''' \
                   '''tokens_per_batch = 1048576\n''' \
                   '''\n''' \
                   '''AdafactorOptimizer.beta1 = 0.0\n''' \
                   '''AdafactorOptimizer.clipping_threshold = 1.0\n''' \
                   '''AdafactorOptimizer.decay_rate = None\n''' \
                   '''AdafactorOptimizer.epsilon1 = 1e-30\n''' \
                   '''AdafactorOptimizer.epsilon2 = 0.001\n''' \
                   '''AdafactorOptimizer.min_dim_size_to_factor = 128\n''' \
                   '''AdafactorOptimizer.multiply_by_parameter_scale = ''' \
                   '''True\n''' \
                   '''Bitransformer.shared_embedding = True\n'''
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.gin_file = os.path.join(self.tmp_dir.name, 'config.gin')
        with open(self.gin_file, 'w') as gin_writer:
            gin_writer.write(test_gin)

    def tearDown(self):
        self.tmp_dir.cleanup()

    def test_read_t5_gin_config_file(self):
        r"""Tests :meth:`~texar.torch.utils.gin.read_t5_gin_config_file`.
        """
        config = read_t5_gin_config_file(self.gin_file)

        expect_config = {'d_ff': 2048,
                         'd_kv': 64,
                         'd_model': 512,
                         'dropout_rate': 0.1,
                         'inputs_length': 512,
                         'num_heads': 8,
                         'num_layers': 6
                         }

        self.assertEqual(config, expect_config)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various connectors.
"""

from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from torch import nn
from torch.distributions.distribution import Distribution

from texar.torch.core import get_activation_fn
from texar.torch.hyperparams import HParams
from texar.torch.modules.connectors.connector_base import ConnectorBase
from texar.torch.utils import nest
from texar.torch.utils import utils
from texar.torch.utils.types import MaybeTuple

__all__ = [
    "ConstantConnector",
    "ForwardConnector",
    "MLPTransformConnector",
    "ReparameterizedStochasticConnector",
    "StochasticConnector",
    # "ConcatConnector"
]

TensorStruct = Union[List[torch.Tensor],
                     Dict[Any, torch.Tensor],
                     MaybeTuple[torch.Tensor]]
OutputSize = MaybeTuple[Union[int, torch.Size]]
ActivationFn = Callable[[torch.Tensor], torch.Tensor]
LinearLayer = Callable[[torch.Tensor], torch.Tensor]


def _assert_same_size(outputs: TensorStruct,
                      output_size: OutputSize):
    r"""Check if outputs match output_size

    Args:
        outputs: A tensor or a (nested) tuple of tensors
        output_size: Can be an ``int``, a ``torch.Size``, or a (nested)
            tuple of ``int`` or ``torch.Size``.
    """
    flat_output_size = nest.flatten(output_size)
    flat_output = nest.flatten(outputs)

    for (output, size) in zip(flat_output, flat_output_size):

        if isinstance(size, torch.Size):
            if output[0].size() != size:
                raise ValueError("The output size does not match"
                                 "the required output_size")
        elif output[0].size()[-1] != size:
            raise ValueError(
                "The output size does not match the required output_size")


def _get_sizes(sizes: List[Any]) -> List[int]:
    r"""

    Args:
        sizes: A list of ``int`` or ``torch.Size``. If each element is of type
            ``torch.Size``, the size is computed by taking the product of the
            shape.

    Returns:
        A list of sizes with ``torch.Size`` replaced by product of its
        individual dimensions
    """
    if isinstance(sizes[0], torch.Size):
        size_list = [np.prod(shape) for shape in sizes]
    else:
        size_list = sizes

    return size_list


def _sum_output_size(output_size: OutputSize) -> int:
    r"""Return sum of all dim values in :attr:`output_size`

    Args:
        output_size: Can be an ``int``, a ``torch.Size``, or a (nested)
            tuple of ``int`` or ``torch.Size``.
    """
    flat_output_size = nest.flatten(output_size)
    size_list = _get_sizes(flat_output_size)
    ret = sum(size_list)
    return ret


def _mlp_transform(inputs: TensorStruct,
                   output_size: OutputSize,
                   linear_layer: Optional[LinearLayer] = None,
                   activation_fn: Optional[ActivationFn] = None) -> Any:
    r"""Transforms inputs through a fully-connected layer that creates
    the output with specified size.

    Args:
        inputs: A Tensor of shape `[batch_size, d1, ..., dn]`, or a (nested)
            tuple of such elements. The dimensions `d1, ..., dn` will be flatten
            and transformed by a dense layer.
        output_size: Can be an ``int``, a ``torch.Size``, or a (nested)
            tuple of ``int`` or ``torch.Size``.
        activation_fn: Activation function applied to the output.

    :returns:
        If :attr:`output_size` is an ``int`` or a ``torch.Size``,
        returns a tensor of shape ``[batch_size, *, output_size]``.
        If :attr:`output_size` is a tuple of ``int`` or ``torch.Size``,
        returns a tuple having the same structure as :attr:`output_size`,
        where each element has the same size as defined in :attr:`output_size`.
    """
    # Flatten inputs
    flat_input = nest.flatten(inputs)
    flat_input = [x.view(-1, x.size(-1)) for x in flat_input]
    concat_input = torch.cat(flat_input, 1)

    # Get output dimension
    flat_output_size = nest.flatten(output_size)

    size_list = _get_sizes(flat_output_size)

    fc_output = concat_input
    if linear_layer is not None:
        fc_output = linear_layer(fc_output)
    if activation_fn is not None:
        fc_output = activation_fn(fc_output)

    flat_output = torch.split(fc_output, size_list, dim=1)
    flat_output = list(flat_output)

    if isinstance(flat_output_size[0], torch.Size):
        flat_output = [torch.reshape(output, (-1,) + shape) for output, shape
                       in zip(flat_output, flat_output_size)]

    output = nest.pack_sequence_as(structure=output_size,
                                   flat_sequence=flat_output)
    return output


class ConstantConnector(ConnectorBase):
    r"""Creates a constant tensor or (nested) tuple of Tensors that
    contains a constant value.

    Args:
        output_size: Size of output **excluding** the batch dimension. For
            example, set :attr:`output_size` to ``dim`` to generate output of
            shape ``[batch_size, dim]``.
            Can be an ``int``, a tuple of ``int``, a ``torch.Size``,
            or a tuple of ``torch.Size``.
            For example, to transform inputs to have decoder state size, set
            :python:`output_size=decoder.state_size`.
            If :attr:`output_size` is a tuple ``(1, 2, 3)``, then the
            output structure will be
            ``([batch_size * 1], [batch_size * 2], [batch_size * 3])``.
            If :attr:`output_size` is ``torch.Size([1, 2, 3])``, then the
            output structure will be ``[batch_size, 1, 2, 3]``.
        hparams (dict, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.
    This connector does not have trainable parameters.

    Example:

        .. code-block:: python

            state_size = (1, 2, 3)
            connector = ConstantConnector(state_size, hparams={"value": 1.})
            one_state = connector(batch_size=64)
            # `one_state` structure: (Tensor_1, Tensor_2, Tensor_3),
            # Tensor_1.size() == torch.Size([64, 1])
            # Tensor_2.size() == torch.Size([64, 2])
            # Tensor_3.size() == torch.Size([64, 3])
            # Tensors are filled with 1.0.
            size = torch.Size([1, 2, 3])
            connector_size = ConstantConnector(size, hparams={"value": 2.})
            size_state = connector_size(batch_size=64)
            # `size_state` structure: Tensor with size [64, 1, 2, 3].
            # Tensor is filled with 2.0.

    """

    def __init__(self,
                 output_size: OutputSize,
                 hparams: Optional[HParams] = None):

        super().__init__(output_size, hparams=hparams)

        self.value = self.hparams.value

    @staticmethod
    def default_hparams() -> dict:
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "value": 0.,
                "name": "constant_connector"
            }

        Here:

        `"value"`: float
            The constant scalar that the output tensor(s) has.
        `"name"`: str
            Name of the connector.
        """
        return {
            "value": 0.,
            "name": "constant_connector"
        }

    def forward(self,  # type: ignore
                batch_size: Union[int, torch.Tensor]) -> Any:
        r"""Creates output tensor(s) that has the given value.

        Args:
            batch_size: An ``int`` or ``int`` scalar tensor, the
                batch size.

        :returns:
            A (structure of) tensor whose structure is the same as
            :attr:`output_size`, with value specified by
            ``value`` or :attr:`hparams`.
        """

        def full_tensor(x):
            if isinstance(x, torch.Size):
                return torch.full((batch_size,) + x, self.value)
            else:
                return torch.full((batch_size, x), self.value)

        output = utils.map_structure(
            full_tensor,
            self._output_size)

        return output


class ForwardConnector(ConnectorBase):
    r"""Transforms inputs to have specified structure.

    Example:

    .. code-block:: python

        state_size = namedtuple('LSTMStateTuple', ['h', 'c'])(256, 256)
        # state_size == LSTMStateTuple(c=256, h=256)
        connector = ForwardConnector(state_size)
        output = connector([tensor_1, tensor_2])
        # output == LSTMStateTuple(c=tensor_1, h=tensor_2)

    Args:
        output_size: Size of output **excluding** the batch dimension. For
            example, set :attr:`output_size` to ``dim`` to generate output of
            shape ``[batch_size, dim]``.
            Can be an ``int``, a tuple of ``int``, a ``torch.Size``, or a
            tuple of ``torch.Size``.
            For example, to transform inputs to have decoder state size, set
            :python:`output_size=decoder.state_size`.
        hparams (dict, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.

    This connector does not have trainable parameters.
    See :meth:`forward` for the inputs and outputs of the connector.
    The input to the connector must have the same structure with
    :attr:`output_size`, or must have the same number of elements and be
    re-packable into the structure of :attr:`output_size`. Note that if input
    is or contains a ``dict`` instance, the keys will be sorted to pack in
    deterministic order (See :func:`~texar.torch.utils.nest.pack_sequence_as`).
    """

    def __init__(self,
                 output_size: OutputSize,
                 hparams: Optional[HParams] = None):
        super().__init__(output_size, hparams=hparams)

    @staticmethod
    def default_hparams() -> dict:
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "name": "forward_connector"
            }

        Here:

        `"name"`: str
            Name of the connector.
        """
        return {
            "name": "forward_connector"
        }

    def forward(self,  # type: ignore
                inputs: TensorStruct) -> Any:
        r"""Transforms inputs to have the same structure as with
        :attr:`output_size`. Values of the inputs are not changed.
        :attr:`inputs` must either have the same structure, or have the same
        number of elements with :attr:`output_size`.

        Args:
            inputs: The input (structure of) tensor to pass forward.

        :returns:
            A (structure of) tensors that re-packs :attr:`inputs` to have
            the specified structure of :attr:`output_size`.
        """
        flat_input = nest.flatten(inputs)
        output = nest.pack_sequence_as(
            self._output_size, flat_input)

        return output


class MLPTransformConnector(ConnectorBase):
    r"""Transforms inputs with an MLP layer and packs the results into the
    specified structure and size.

    Example:

        .. code-block:: python

            cell = LSTMCell(num_units=256)
            # cell.state_size == LSTMStateTuple(c=256, h=256)
            connector = MLPTransformConnector(cell.state_size)
            inputs = torch.zeros([64, 10])
            output = connector(inputs)
            # output == LSTMStateTuple(c=tensor_of_shape_(64, 256),
            #                          h=tensor_of_shape_(64, 256))

        .. code-block:: python

            ## Use to connect encoder and decoder with different state size
            encoder = UnidirectionalRNNEncoder(...)
            _, final_state = encoder(inputs=...)
            decoder = BasicRNNDecoder(...)
            connector = MLPTransformConnector(decoder.state_size)
            _ = decoder(
                initial_state=connector(final_state),
                ...)

    Args:
        output_size: Size of output **excluding** the batch dimension. For
            example, set :attr:`output_size` to ``dim`` to generate output of
            shape ``[batch_size, dim]``.
            Can be an ``int``, a tuple of ``int``, a ``torch.Size``,
            or a tuple of ``torch.Size``.
            For example, to transform inputs to have decoder state size, set
            :python:`output_size=decoder.state_size`.
        linear_layer_dim (int): Value of final dim of the input tensors i.e. the
            input dim of the mlp linear layer.
        hparams (dict, optional): Hyperparameters. Missing hyperparameter will
            be set to default values. See :meth:`default_hparams` for the
            hyperparameter structure and default values.

    The input to the connector can have arbitrary structure and size.
    """

    def __init__(self,
                 output_size: OutputSize,
                 linear_layer_dim: int,
                 hparams: Optional[HParams] = None):
        super().__init__(output_size, hparams=hparams)
        self._linear_layer = nn.Linear(
            linear_layer_dim, _sum_output_size(output_size))
        self._activation_fn = get_activation_fn(
            self.hparams.activation_fn)

    @staticmethod
    def default_hparams() -> dict:
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "activation_fn": "texar.torch.core.layers.identity",
                "name": "mlp_connector"
            }

        Here:

        `"activation_fn"`: str or callable
            The activation function applied to the outputs of the MLP
            transformation layer. Can
            be a function, or its name or module path.
        `"name"`: str
            Name of the connector.
        """
        return {
            "activation_fn": "texar.torch.core.layers.identity",
            "name": "mlp_connector"
        }

    def forward(self,  # type: ignore
                inputs: TensorStruct) -> Any:
        r"""Transforms inputs with an MLP layer and packs the results to have
        the same structure as specified by :attr:`output_size`.

        Args:
            inputs: Input (structure of) tensors to be transformed. Must be a
                tensor of shape ``[batch_size, ...]`` or a (nested)
                tuple of such Tensors. That is, the first dimension of
                (each) tensor must be the batch dimension.

        :returns:
            A tensor or a (nested) tuple of tensors of the same structure of
            :attr:`output_size`.
        """

        output = _mlp_transform(
            inputs, self._output_size,
            self._linear_layer, self._activation_fn)

        return output


class ReparameterizedStochasticConnector(ConnectorBase):
    r"""Samples from a distribution with reparameterization trick, and
    transforms samples into specified size.
    Reparameterization allows gradients to be back-propagated through the
    stochastic samples. Used in, e.g., Variational Autoencoders (VAEs).

    Example:

        .. code-block:: python

            # Initialized without num_samples
            cell = LSTMCell(num_units=256)
            # cell.state_size == LSTMStateTuple(c=256, h=256)
            mu = torch.zeros([16, 100])
            var = torch.ones([100])

            connector = ReparameterizedStochasticConnector(
                cell.state_size,
                mlp_input_size=mu.size()[-1],
                distribution="MultivariateNormal",
                distribution_kwargs={
                    "loc": mu,
                    "scale_tril": torch.diag(var)})
            output, sample = connector()
            # output == LSTMStateTuple(c=tensor_of_shape_(16, 256),
            #                          h=tensor_of_shape_(16, 256))
            # sample == Tensor([16, 100])

            output_, sample_ = connector(num_samples=4)
            # output_ == LSTMStateTuple(c=tensor_of_shape_(4, 16, 256),
            #                           h=tensor_of_shape_(4, 16, 256))
            # sample == Tensor([4, 16, 100])

    Args:
        output_size: Size of output **excluding** the batch dimension. For
            example, set ``output_size`` to ``dim`` to generate output of
            shape ``[batch_size, dim]``.
            Can be an ``int``, a tuple of ``int``, a ``torch.Size``, or
            a tuple of ``torch.Size``.
            For example, to transform inputs to have decoder state size, set
            :python:`output_size=decoder.state_size`.
        mlp_input_size: Size of MLP transfer process input, which is equal to
            the distribution result size **excluding** the batch dimension,
            Can be ``int`` or ``torch.Size`` or a tuple of ``int``.
        distribution: A instance or name ``str`` of subclass of
            :torch:`distributions.distribution.Distribution`,
            Can be a distribution class instance or ``str``.
        distribution_kwargs (dict, optional): ``dict`` of keyword arguments
            for the :attr:`distribution`. Its keys are `str`, which are names
            of keyword arguments; Its values are corresponding values for each
            argument.
        hparams (dict, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.
    """

    def __init__(self,
                 output_size: OutputSize,
                 mlp_input_size: Union[torch.Size, MaybeTuple[int], int],
                 distribution: Union[Distribution, str] = 'MultivariateNormal',
                 distribution_kwargs: Optional[Dict[str, Any]] = None,
                 hparams: Optional[HParams] = None):
        super().__init__(output_size, hparams=hparams)
        if distribution_kwargs is None:
            distribution_kwargs = {}
        self._dstr_type = distribution
        self._dstr_kwargs = distribution_kwargs

        for dstr_attr, dstr_val in distribution_kwargs.items():
            if isinstance(dstr_val, torch.Tensor):
                dstr_param = nn.Parameter(dstr_val)
                distribution_kwargs[dstr_attr] = dstr_param
                self.register_parameter(dstr_attr, dstr_param)
        if isinstance(mlp_input_size, int):
            input_feature = mlp_input_size
        else:
            input_feature = np.prod(mlp_input_size)
        self._linear_layer = nn.Linear(
            input_feature, _sum_output_size(output_size))

        self._activation_fn = get_activation_fn(
            self.hparams.activation_fn)

    @staticmethod
    def default_hparams() -> dict:
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "activation_fn": "texar.torch.core.layers.identity",
                "name": "reparameterized_stochastic_connector"
            }

        Here:

        `"activation_fn"`: str
            The activation function applied to the outputs of the MLP
            transformation layer. Can be a function, or its name or module path.
        `"name"`: str
            Name of the connector.
        """
        return {
            "activation_fn": "texar.torch.core.layers.identity",
            "name": "reparameterized_stochastic_connector"
        }

    def forward(self,  # type: ignore
                num_samples: Optional[Union[int, torch.Tensor]] = None,
                transform: bool = True) -> Tuple[Any, Any]:
        r"""Samples from a distribution and optionally performs transformation
        with an MLP layer.
        The distribution must be reparameterizable, i.e.,
        :python:`Distribution.has_rsample == True`.

        Args:
            num_samples (optional): An ``int`` or ``int`` tensor.
                Number of samples to generate. If not given,
                generate a single sample. Note that if batch size has
                already been included in :attr:`distribution`'s dimensionality,
                :attr:`num_samples` should be left as ``None``.
            transform (bool): Whether to perform MLP transformation of the
                distribution samples. If ``False``, the structure/shape of a
                sample must match :attr:`output_size`.


        :returns:
            A tuple (:attr:`output`, :attr:`sample`), where

            - output: A tensor or a (nested) tuple of Tensors with
              the same structure and size of :attr:`output_size`.
              The batch dimension equals :attr:`num_samples` if specified,
              or is determined by the distribution dimensionality.
              If :attr:`transform` is `False`, it will be
              equal to :attr:`sample`.
            - sample: The sample from the distribution, prior to transformation.

            Otherwise, returns a tensor :attr:`sample`, where
            - sample: The sample from the distribution, prior to transformation.

        Raises:
            ValueError: If distribution is not reparameterizable.
            ValueError: The output does not match :attr:`output_size`.
        """
        if isinstance(self._dstr_type, str):
            dstr: Distribution = utils.check_or_get_instance(
                self._dstr_type, self._dstr_kwargs,
                ["torch.distributions", "texar.torch.custom"])
        else:
            dstr = self._dstr_type

        if not dstr.has_rsample:
            raise ValueError("Distribution should be reparameterizable")

        if num_samples:
            sample = dstr.rsample([num_samples])
        else:
            sample = dstr.rsample()

        if transform:
            output = _mlp_transform(
                sample,
                self._output_size,
                self._linear_layer,
                self._activation_fn)
            _assert_same_size(output, self._output_size)

        else:
            output = sample

        return output, sample


class StochasticConnector(ConnectorBase):
    r"""Samples from a distribution and transforms samples into specified size.
    The connector is the same as
    :class:`~texar.torch.modules.ReparameterizedStochasticConnector`, except
    that here reparameterization is disabled, and thus the gradients cannot be
    back-propagated through the stochastic samples.

    Args:
        output_size: Size of output **excluding** the batch dimension. For
            example, set ``output_size`` to ``dim`` to generate output of
            shape ``[batch_size, dim]``.
            Can be an ``int``, a tuple of ``int``, a torch.Size, or a tuple of
            torch.Size.
            For example, to transform inputs to have decoder state size, set
            :python:`output_size=decoder.state_size`.
        mlp_input_size: Size of MLP transfer process input, which is equal to
            the distribution result size **excluding** the batch dimension,
            Can be ``int`` or ``torch.Size`` or a tuple of ``int``.
        distribution: A instance of subclass of
            :torch:`distributions.distribution.Distribution`,
            Can be a class, its name or module path, or a class instance.
            The :attr:`distribution` should not be reparameterizable.
        distribution_kwargs (dict, optional): ``dict`` of keyword arguments
            for the :attr:`distribution`. Its keys are `str`, which are names
            of keyword arguments; Its values are corresponding values for each
            argument.
        hparams (dict, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.
    """

    def __init__(self,
                 output_size: OutputSize,
                 mlp_input_size: Union[torch.Size, MaybeTuple[int], int],
                 distribution: Union[Distribution, str] = 'MultivariateNormal',
                 distribution_kwargs: Optional[Dict[str, Any]] = None,
                 hparams: Optional[HParams] = None):
        super().__init__(output_size, hparams=hparams)
        if distribution_kwargs is None:
            distribution_kwargs = {}
        self._dstr_kwargs = distribution_kwargs
        if isinstance(distribution, str):
            self._dstr: Distribution = utils.check_or_get_instance(
                distribution, self._dstr_kwargs,
                ["torch.distributions", "texar.torch.custom"])
        else:
            self._dstr = distribution

        if self._dstr.has_rsample:
            raise ValueError("Distribution should not be reparameterizable")

        if isinstance(mlp_input_size, int):
            input_feature = mlp_input_size
        else:
            input_feature = np.prod(mlp_input_size)
        self._linear_layer = nn.Linear(
            input_feature, _sum_output_size(output_size))

        self._activation_fn = get_activation_fn(
            self.hparams.activation_fn)

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "activation_fn": "texar.torch.core.layers.identity",
                "name": "stochastic_connector"
            }

        Here:

        `"activation_fn"`: str
            The activation function applied to the outputs of the MLP
            transformation layer. Can
            be a function, or its name or module path.
        `"name"`: str
            Name of the connector.
        """
        return {
            "activation_fn": "texar.torch.core.layers.identity",
            "name": "stochastic_connector"
        }

    def forward(self,  # type: ignore
                num_samples: Optional[Union[int, torch.Tensor]] = None,
                transform: bool = False) -> Any:
        r"""Samples from a distribution and optionally performs transformation
        with an MLP layer.
        The inputs and outputs are the same as
        :class:`~texar.torch.modules.ReparameterizedStochasticConnector` except
        that the distribution does not need to be reparameterizable, and
        gradient cannot be back-propagate through the samples.

        Args:
            num_samples (optional): An ``int`` or ``int`` tensor.
                Number of samples to generate. If not given,
                generate a single sample. Note that if batch size has
                already been included in :attr:`distribution`'s dimensionality,
                :attr:`num_samples` should be left as ``None``.
            transform (bool): Whether to perform MLP transformation of the
                distribution samples. If ``False``, the structure/shape of a
                sample must match :attr:`output_size`.

        :returns:
            A tuple (:attr:`output`, :attr:`sample`), where

            - output: A tensor or a (nested) tuple of Tensors with
              the same structure and size of :attr:`output_size`.
              The batch dimension equals :attr:`num_samples` if specified,
              or is determined by the distribution dimensionality.
              If :attr:`transform` is `False`, it will be
              equal to :attr:`sample`.
            - sample: The sample from the distribution, prior to transformation.

        Raises:

            ValueError: If distribution can be reparameterizable.
            ValueError: The output does not match :attr:`output_size`.
        """

        if num_samples:
            sample = self._dstr.sample([num_samples])
        else:
            sample = self._dstr.sample()

        if self._dstr.event_shape == []:
            sample = torch.reshape(
                input=sample, shape=sample.size() + torch.Size([1]))

        # Disable gradients through samples
        sample = sample.detach().float()
        if transform:
            output = _mlp_transform(
                sample,
                self._output_size,
                self._linear_layer,
                self._activation_fn)
            _assert_same_size(output, self._output_size)

        else:
            output = sample

        return output, sample

# class ConcatConnector(ConnectorBase):
#    r"""Concatenates multiple connectors into one connector. Used in, e.g.,
#    semi-supervised variational autoencoders, disentangled representation
#    learning, and other models.
#
#    Args:
#        output_size: Size of output excluding the batch dimension (eg.
#            :attr:`output_size = p` if :attr:`output.shape` is :attr:`[N, p]`).
#            Can be an int, a tuple of int, a torch.Size, or a tuple of
#            torch.Size.
#            For example, to transform to decoder state size, set
#            `output_size=decoder.cell.state_size`.
#        hparams (dict): Hyperparameters of the connector.
#    """
#
#    def __init__(self, output_size, hparams=None):
#        super().__init__(self, output_size, hparams)
#
#    @staticmethod
#    def default_hparams():
#        r"""Returns a dictionary of hyperparameters with default values.
#
#        Returns:
#
#            .. code-block:: python
#
#                {
#                    "activation_fn": "texar.torch.core.layers.identity",
#                    "name": "concat_connector"
#                }
#
#            Here:
#
#            `"activation_fn"`: (str or callable)
#                The name or full path to the activation function applied to
#                the outputs of the MLP layer. The activation functions can be:
#
#                - Built-in activation functions defined in :
#                - User-defined activation functions in `texar.torch.custom`.
#                - External activation functions. Must provide the full path, \
#                  e.g., "my_module.my_activation_fn".
#
#                The default value is :attr:`"identity"`, i.e., the MLP
#                transformation is linear.
#
#            `"name"`: str
#                Name of the connector.
#
#                The default value is "concat_connector".
#        """
#        return {
#            "activation_fn": "texar.torch.core.layers.identity",
#            "name": "concat_connector"
#        }
#
#    def forward(self, connector_inputs, transform=True):
#        r"""Concatenate multiple input connectors
#
#        Args:
#            connector_inputs: a list of connector states
#            transform (bool): If `True`, then the output are automatically
#                transformed to match :attr:`output_size`.
#
#        Returns:
#            A Tensor or a (nested) tuple of Tensors of the same structure of
#            the decoder state.
#        """
#        connector_inputs = [connector.float()
#                            for connector in connector_inputs]
#        output = torch.cat(connector_inputs, dim=1)
#
#        if transform:
#            fn_modules = ['texar.torch.custom', 'torch', 'torch.nn']
#            activation_fn = utils.get_function(self.hparams.activation_fn,
#                                         fn_modules)
#            output, linear_layer = _mlp_transform(
#               output, self._output_size, activation_fn)
#            self._linear_layers.append(linear_layer)
#        _assert_same_size(output, self._output_size)
#
#        self._add_internal_trainable_variables()
#        self._built = True
#
#        return output



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for XLNet encoder.
"""
import unittest

import torch

from texar.torch.modules.encoders.xlnet_encoder import XLNetEncoder
from texar.torch.utils.test import pretrained_test


class XLNetEncoderTest(unittest.TestCase):
    r"""Tests :class:`~texar.torch.modules.XLNetEncoder` class.
    """

    def setUp(self) -> None:
        self.batch_size = 2
        self.max_length = 3
        self.inputs = torch.zeros(
            self.batch_size, self.max_length, dtype=torch.long)

    @pretrained_test
    def test_model_loading(self):
        r"""Tests model loading functionality."""
        for pretrained_model_name in XLNetEncoder.available_checkpoints():
            encoder = XLNetEncoder(pretrained_model_name=pretrained_model_name)
            _ = encoder(self.inputs)

    @pretrained_test
    def test_hparams(self):
        r"""Tests the priority of the encoder arch parameter.
        """
        # case 1: set "pretrained_mode_name" by constructor argument
        hparams = {
            "pretrained_model_name": "xlnet-large-cased",
        }
        encoder = XLNetEncoder(pretrained_model_name="xlnet-base-cased",
                               hparams=hparams)
        self.assertEqual(encoder.hparams.num_layers, 12)
        _, _ = encoder(self.inputs)

        # case 2: set "pretrained_mode_name" by hparams
        hparams = {
            "pretrained_model_name": "xlnet-large-cased",
            "num_layers": 6,
        }
        encoder = XLNetEncoder(hparams=hparams)
        self.assertEqual(encoder.hparams.num_layers, 24)
        _, _ = encoder(self.inputs)

        # case 3: set to None in both hparams and constructor argument
        hparams = {
            "pretrained_model_name": None,
            "num_layers": 6,
        }
        encoder = XLNetEncoder(hparams=hparams)
        self.assertEqual(encoder.hparams.num_layers, 6)
        _, _ = encoder(self.inputs)

        # case 4: using default hparams
        encoder = XLNetEncoder()
        self.assertEqual(encoder.hparams.num_layers, 12)
        _, _ = encoder(self.inputs)

    @pretrained_test
    def test_trainable_variables(self):
        r"""Tests the functionality of automatically collecting trainable
        variables.
        """
        # case 1: xlnet base
        encoder = XLNetEncoder()
        self.assertEqual(len(encoder.trainable_variables), 182)
        _, _ = encoder(self.inputs)

        # Case 2: xlnet large
        hparams = {
            "pretrained_model_name": "xlnet-large-cased",
        }
        encoder = XLNetEncoder(hparams=hparams)
        self.assertEqual(len(encoder.trainable_variables), 362)
        _, _ = encoder(self.inputs)

        # case 3: self-designed bert
        hparams = {
            "num_layers": 6,
            "pretrained_model_name": None,
        }
        encoder = XLNetEncoder(hparams=hparams)
        self.assertEqual(len(encoder.trainable_variables), 92)
        _, _ = encoder(self.inputs)

    def test_encode(self):
        r"""Tests encoding.
        """
        # case 1: xlnet base
        hparams = {
            "pretrained_model_name": None,
        }
        encoder = XLNetEncoder(hparams=hparams)

        inputs = torch.randint(32000, (self.batch_size, self.max_length))
        outputs, new_memory = encoder(inputs)

        self.assertEqual(
            outputs.shape,
            torch.Size([self.batch_size, self.max_length, encoder.output_size]))
        self.assertEqual(new_memory, None)

        # case 2: self-designed xlnet
        hparams = {
            'pretrained_model_name': None,
            'untie_r': True,
            'num_layers': 6,
            'mem_len': 0,
            'reuse_len': 0,
            'num_heads': 8,
            'hidden_dim': 32,
            'head_dim': 64,
            'dropout': 0.1,
            'attention_dropout': 0.1,
            'use_segments': True,
            'ffn_inner_dim': 256,
            'activation': 'gelu',
            'vocab_size': 32000,
            'max_seq_length': 128,
            'initializer': None,
            'name': "xlnet_encoder",
        }
        encoder = XLNetEncoder(hparams=hparams)
        outputs, new_memory = encoder(inputs)

        self.assertEqual(
            outputs.shape,
            torch.Size([self.batch_size, self.max_length, encoder.output_size]))
        self.assertEqual(new_memory, None)

    def test_soft_ids(self):
        r"""Tests soft ids.
        """
        hparams = {
            "pretrained_model_name": None,
        }
        encoder = XLNetEncoder(hparams=hparams)

        inputs = torch.rand(self.batch_size, self.max_length, 32000)
        outputs, new_memory = encoder(inputs)

        self.assertEqual(
            outputs.shape,
            torch.Size([self.batch_size, self.max_length, encoder.output_size]))
        self.assertEqual(new_memory, None)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for attention mechanism utils.
"""

import unittest

import torch

from texar.torch.core.attention_mechanism import (
    maybe_mask_score, prepare_memory, safe_cumprod)
from texar.torch.core.attention_mechanism_utils import hardmax, sparsemax


class AttentionMechanismUtilsTest(unittest.TestCase):
    r"""Tests attention mechanism utils.
    """

    def test_prepare_memory(self):
        r"""Tests `texar.torch.core.attention_mechanism_utils.prepare_memory`.
        """
        memory = torch.tensor([[[1, 2, 3], [4, 5, 6], [7, 8, 9], [9, 8, 7]],
                               [[3, 2, 1], [6, 5, 4], [9, 8, 7], [4, 5, 6]]])
        memory_sequence_length = torch.tensor([3, 2])
        masked_memory = prepare_memory(memory, memory_sequence_length)

        expected_memory = [[[1, 2, 3], [4, 5, 6], [7, 8, 9], [0, 0, 0]],
                           [[3, 2, 1], [6, 5, 4], [0, 0, 0], [0, 0, 0]]]

        self.assertEqual(masked_memory.tolist(), expected_memory)

    def test_maybe_mask_score(self):
        r"""Tests `texar.torch.core.attention_mechanism_utils.maybe_mask_score`.
        """
        score = torch.tensor([[1, 2, 3],
                              [4, 5, 6],
                              [7, 8, 9]])
        score_mask_value = torch.tensor(-1)
        memory_sequence_length = torch.tensor([1, 2, 3])
        masked_score = maybe_mask_score(
            score, score_mask_value, memory_sequence_length)

        expected_score = [[1, -1, -1], [4, 5, -1], [7, 8, 9]]

        self.assertEqual(masked_score.tolist(), expected_score)

    def test_hardmax(self):
        r"""Tests `texar.torch.core.attention_mechanism_utils.hardmax`.
        """
        logits = torch.tensor([[[1, 2, 3], [4, 5, 6], [7, 8, 9], [9, 8, 7]],
                               [[3, 2, 1], [6, 5, 4], [9, 8, 7], [4, 5, 6]]])
        outputs = hardmax(logits)

        expected_outputs = [[[0, 0, 1], [0, 0, 1], [0, 0, 1], [1, 0, 0]],
                            [[1, 0, 0], [1, 0, 0], [1, 0, 0], [0, 0, 1]]]

        self.assertEqual(outputs.tolist(), expected_outputs)

    def test_safe_cumprod(self):
        r"""Tests `texar.torch.core.attention_mechanism_utils.safe_cumprod`.
        """
        x = torch.tensor([0.1, 0.2, 0.3, 0.4, 0.5])
        outputs = safe_cumprod(x, dim=0)

        expected_outputs = [0.1, 0.02, 0.006, 0.0024, 0.0012]

        outputs = outputs.tolist()
        for i in range(5):
            self.assertAlmostEqual(outputs[i], expected_outputs[i])

    def test_sparsemax(self):
        r"""Tests `texar.torch.core.attention_mechanism_utils.sparsemax`.
        """
        logits = torch.tensor([[1.2, 3.2, 0.1, 4.3, 5.0],
                               [3.3, 4.6, 9.5, 5.4, 8.7]])
        outputs = sparsemax(logits)

        expected_outputs = [[0.0, 0.0, 0.0, 0.1500001, 0.8499999],
                            [0.0, 0.0, 0.8999996, 0.0, 0.09999943]]

        outputs = outputs.tolist()
        for i in range(2):
            for j in range(5):
                self.assertAlmostEqual(outputs[i][j],
                                       expected_outputs[i][j],
                                       places=5)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for beam search.
"""

import unittest

import torch
import numpy as np

from texar.torch.utils import beam_search


class BeamSearchTest(unittest.TestCase):
    r"""Tests beam_search.
    """

    def testShapes(self):
        batch_size = 2
        beam_size = 3
        vocab_size = 4
        decode_length = 10

        initial_ids = torch.tensor([0, 0], dtype=torch.int64)

        def symbols_to_logits(_):
            # Just return random logits
            return torch.rand(batch_size * beam_size, vocab_size)

        final_ids, final_probs = beam_search.beam_search(
            symbols_to_logits_fn=symbols_to_logits,
            initial_ids=initial_ids,
            beam_size=beam_size,
            decode_length=decode_length,
            vocab_size=vocab_size,
            alpha=0.0,
            eos_id=1)

        self.assertEqual(final_ids.shape[1], beam_size)
        self.assertEqual(final_probs.shape, torch.Size([batch_size, beam_size]))

    def testComputeTopkScoresAndSeq(self):
        batch_size = 2
        beam_size = 3

        sequences = torch.tensor([[[2, 3], [4, 5], [6, 7], [19, 20]],
                                  [[8, 9], [10, 11], [12, 13], [80, 17]]],
                                 dtype=torch.int64)

        scores = torch.tensor([[-0.1, -2.5, 0., -1.5],
                               [-100., -5., -0.00789, -1.34]])

        flags = torch.tensor([[True, False, False, True],
                              [False, False, False, True]])

        topk_seq, topk_scores, topk_flags, _ = (
            beam_search.compute_topk_scores_and_seq(sequences=sequences,
                                                    scores=scores,
                                                    scores_to_gather=scores,
                                                    flags=flags,
                                                    beam_size=beam_size,
                                                    batch_size=batch_size))

        exp_seq = [[[6, 7], [2, 3], [19, 20]], [[12, 13], [80, 17], [10, 11]]]
        exp_scores = [[0., -0.1, -1.5], [-0.00789, -1.34, -5.]]
        exp_flags = [[False, True, True], [False, True, False]]

        self.assertEqual(topk_seq.tolist(), exp_seq)
        topk_scores = topk_scores.tolist()
        for i in range(2):
            for j in range(3):
                self.assertAlmostEqual(topk_scores[i][j], exp_scores[i][j])
        self.assertEqual(topk_flags.tolist(), exp_flags)

    def testGreedyBatchOne(self):
        batch_size = 1
        beam_size = 1
        vocab_size = 2
        decode_length = 3

        initial_ids = torch.tensor([0] * batch_size, dtype=torch.int64)

        # Test that beam search finds the most probable sequence.
        # These probabilities represent the following search
        #
        #               G0 (0)
        #                  / \
        #                /     \
        #              /         \
        #            /             \
        #         0(0.7)          1(0.3)
        #           / \
        #          /   \
        #         /     \
        #     0(0.4) 1(0.6)
        #        /\
        #       /  \
        #      /    \
        #    0(0.5) 1(0.5)
        # and the following decoding probabilities
        # 0000 - 0.7 * 0.4  * 0.1
        # 0001 - 0.7 * 0.4  * 0.9
        # 001 - 0.7 * 0.6 (Best)
        # 01 = 0.3
        #
        # 001 is the most likely sequence under these probabilities.
        probabilities = torch.tensor([[[0.7, 0.3]], [[0.4, 0.6]], [[0.5, 0.5]]])

        def symbols_to_logits(ids):
            pos = ids.shape[1]
            logits = torch.log(probabilities[pos - 1, :]).type(torch.float)
            return logits

        final_ids, final_probs = beam_search.beam_search(
            symbols_to_logits_fn=symbols_to_logits,
            initial_ids=initial_ids,
            beam_size=beam_size,
            decode_length=decode_length,
            vocab_size=vocab_size,
            alpha=0.0,
            eos_id=1)

        exp_ids = [[[0, 0, 1]]]
        exp_probs = [[0.7 * 0.6]]

        self.assertEqual(final_ids.tolist(), exp_ids)
        self.assertAlmostEqual(np.exp(final_probs).tolist()[0][0],
                               exp_probs[0][0])

    def testNotGreedyBeamTwoWithStopEarly(self):
        batch_size = 1
        beam_size = 2
        vocab_size = 3
        decode_length = 10

        initial_ids = torch.tensor([0] * batch_size, dtype=torch.int64)
        probabilities = torch.tensor([[[0.1, 0.1, 0.8], [0.1, 0.1, 0.8]],
                                      [[0.4, 0.5, 0.1], [0.2, 0.4, 0.4]],
                                      [[0.05, 0.9, 0.05], [0.4, 0.4, 0.2]]])

        def symbols_to_logits(ids):
            pos = ids.shape[1]
            logits = torch.log(probabilities[pos - 1, :]).type(torch.float)
            return logits

        final_ids, final_probs = beam_search.beam_search(
            symbols_to_logits_fn=symbols_to_logits,
            initial_ids=initial_ids,
            beam_size=beam_size,
            decode_length=decode_length,
            vocab_size=vocab_size,
            alpha=0.0,
            eos_id=1,
            stop_early=True)  # default value, but just to make this explicit

        # given stop_early = True, the only 'assurance' is w.r.t. the first beam
        # (i.e., other beams may not even be completed)
        # so, we check only the first beam
        first_beam = final_ids[:, 0]
        first_probs = final_probs[:, 0]

        exp_ids = [[0, 2, 1]]
        exp_probs = [0.8 * 0.5]

        self.assertEqual(first_beam.tolist(), exp_ids)
        self.assertAlmostEqual(np.exp(first_probs).tolist()[0], exp_probs[0])

    def testNotGreedyBeamTwoWithoutStopEarly(self):
        batch_size = 1
        beam_size = 2
        vocab_size = 3
        decode_length = 3

        initial_ids = torch.tensor([0] * batch_size, dtype=torch.int64)
        probabilities = torch.tensor([[[0.1, 0.1, 0.8], [0.1, 0.1, 0.8]],
                                      [[0.4, 0.5, 0.1], [0.2, 0.4, 0.4]],
                                      [[0.05, 0.9, 0.05], [0.4, 0.4, 0.2]]])

        def symbols_to_logits(ids):
            pos = ids.shape[1]
            logits = torch.log(probabilities[pos - 1, :]).type(torch.float)
            return logits

        final_ids, final_probs = beam_search.beam_search(
            symbols_to_logits_fn=symbols_to_logits,
            initial_ids=initial_ids,
            beam_size=beam_size,
            decode_length=decode_length,
            vocab_size=vocab_size,
            alpha=0.0,
            eos_id=1,
            stop_early=False)

        # given stop_early = False, the algorithm will return all the beams
        # so we can test all of them here

        exp_ids = [[[0, 2, 1, 0], [0, 2, 0, 1]]]
        exp_probs = [[0.8 * 0.5, 0.8 * 0.4 * 0.9]]

        self.assertEqual(final_ids.tolist(), exp_ids)
        self.assertAlmostEqual(np.exp(final_probs).tolist()[0][0],
                               exp_probs[0][0])
        self.assertAlmostEqual(np.exp(final_probs).tolist()[0][1],
                               exp_probs[0][1])

    def testGreedyWithCornerCase(self):
        batch_size = 1
        beam_size = 1
        vocab_size = 3
        decode_length = 2

        initial_ids = torch.tensor([0] * batch_size, dtype=torch.int64)
        probabilities = torch.tensor([[0.2, 0.1, 0.7], [0.4, 0.1, 0.5]])

        def symbols_to_logits(ids):
            pos = ids.shape[1]
            logits = torch.log(probabilities[pos - 1, :]).type(torch.float)
            return logits

        final_ids, final_probs = beam_search.beam_search(
            symbols_to_logits_fn=symbols_to_logits,
            initial_ids=initial_ids,
            beam_size=beam_size,
            decode_length=decode_length,
            vocab_size=vocab_size,
            alpha=0.0,
            eos_id=1)

        exp_ids = [[[0, 2, 2]]]
        exp_probs = [[0.7 * 0.5]]

        self.assertEqual(final_ids.tolist(), exp_ids)
        self.assertAlmostEqual(np.exp(final_probs).tolist()[0][0],
                               exp_probs[0][0])

    def testNotGreedyBatchTwoBeamTwoWithAlpha(self):
        batch_size = 2
        beam_size = 2
        vocab_size = 3
        decode_length = 3

        initial_ids = torch.tensor([0] * batch_size, dtype=torch.int64)
        # Probabilities for position * batch * beam * vocab
        # Probabilities have been set such that with alpha = 3.5, the less
        # probable but longer sequence will have a better score than the
        # shorter sequence with higher log prob in batch 1, and the order will
        # be reverse in batch 2. That is, the shorter sequence will still have
        # a higher score in spite of the length penalty
        probabilities = torch.tensor([[[[0.1, 0.1, 0.8], [0.1, 0.1, 0.8]],
                                       [[0.1, 0.1, 0.8], [0.1, 0.1, 0.8]]],
                                      [[[0.4, 0.5, 0.1], [0.2, 0.4, 0.4]],
                                       [[0.3, 0.6, 0.1], [0.2, 0.4, 0.4]]],
                                      [[[0.05, 0.9, 0.05], [0.4, 0.4, 0.2]],
                                       [[0.05, 0.9, 0.05], [0.4, 0.4, 0.2]]]])

        def symbols_to_logits(ids):
            pos = ids.shape[1]
            logits = torch.log(probabilities[pos - 1, :]).type(torch.float)
            return logits

        final_ids, final_probs = beam_search.beam_search(
            symbols_to_logits_fn=symbols_to_logits,
            initial_ids=initial_ids,
            beam_size=beam_size,
            decode_length=decode_length,
            vocab_size=vocab_size,
            alpha=3.5,
            eos_id=1)

        exp_ids = [[[0, 2, 0, 1], [0, 2, 1, 0]], [[0, 2, 1, 0], [0, 2, 0, 1]]]
        exp_probs = [[np.log(0.8 * 0.4 * 0.9) / (8. / 6.)**3.5,
                      np.log(0.8 * 0.5) / (7. / 6.)**3.5],
                     [np.log(0.8 * 0.6) / (7. / 6.)**3.5,
                      np.log(0.8 * 0.3 * 0.9) / (8. / 6.)**3.5]]

        self.assertEqual(final_ids.tolist(), exp_ids)
        for i in range(2):
            for j in range(2):
                self.assertAlmostEqual(final_probs.tolist()[i][j],
                                       exp_probs[i][j])

    def testNotGreedyBeamTwoWithAlpha(self):
        batch_size = 1
        beam_size = 2
        vocab_size = 3
        decode_length = 3

        initial_ids = torch.tensor([0] * batch_size, dtype=torch.int64)
        # Probabilities for position * batch * beam * vocab
        # Probabilities have been set such that with alpha = 3.5, the less
        # probable but longer sequence will have a better score that the
        # shorter sequence with higher log prob.
        probabilities = torch.tensor([[[0.1, 0.1, 0.8], [0.1, 0.1, 0.8]],
                                      [[0.4, 0.5, 0.1], [0.2, 0.4, 0.4]],
                                      [[0.05, 0.9, 0.05], [0.4, 0.4, 0.2]]])

        def symbols_to_logits(ids):
            pos = ids.shape[1]
            logits = torch.log(probabilities[pos - 1, :]).type(torch.float)
            return logits

        # Disable early stopping
        final_ids, final_probs = beam_search.beam_search(
            symbols_to_logits_fn=symbols_to_logits,
            initial_ids=initial_ids,
            beam_size=beam_size,
            decode_length=decode_length,
            vocab_size=vocab_size,
            alpha=3.5,
            eos_id=1)

        exp_ids = [[[0, 2, 0, 1], [0, 2, 1, 0]]]
        exp_probs = [[np.log(0.8 * 0.4 * 0.9) / (8. / 6.)**3.5,
                      np.log(0.8 * 0.5) / (7. / 6.)**3.5]]

        self.assertEqual(final_ids.tolist(), exp_ids)
        self.assertAlmostEqual(final_probs.tolist()[0][0], exp_probs[0][0])
        self.assertAlmostEqual(final_probs.tolist()[0][1], exp_probs[0][1])

    def testStates(self):
        batch_size = 1
        beam_size = 1
        vocab_size = 2
        decode_length = 3

        initial_ids = torch.tensor([0] * batch_size, dtype=torch.int64)
        probabilities = torch.tensor([[[0.7, 0.3]], [[0.4, 0.6]], [[0.5, 0.5]]])

        expected_states = torch.tensor([[[0.]], [[1.]]])

        def symbols_to_logits(ids, states):
            pos = ids.shape[1]
            logits = torch.log(probabilities[pos - 1, :]).type(torch.float)
            states["state"] += 1
            return logits, states

        states = {
            "state": torch.zeros(batch_size, 1),
        }

        final_ids, _ = beam_search.beam_search(
            symbols_to_logits_fn=symbols_to_logits,
            initial_ids=initial_ids,
            beam_size=beam_size,
            decode_length=decode_length,
            vocab_size=vocab_size,
            alpha=0.0,
            eos_id=1,
            states=states)

    def testStateBeamTwo(self):
        batch_size = 1
        beam_size = 2
        vocab_size = 3
        decode_length = 3

        initial_ids = torch.tensor([0] * batch_size, dtype=torch.int64)
        probabilities = torch.tensor([[[0.1, 0.1, 0.8], [0.1, 0.1, 0.8]],
                                      [[0.4, 0.5, 0.1], [0.2, 0.4, 0.4]],
                                      [[0.05, 0.9, 0.05], [0.4, 0.4, 0.2]]])

        # The top beam is always selected so we should see the top beam's state
        # at each position, which is the one that getting 3 added to it each
        # step.
        expected_states = torch.tensor([[[0.], [0.]], [[3.], [3.]], [[6.],
                                                                     [6.]]])

        def symbols_to_logits(ids, states):
            pos = ids.shape[1]
            logits = torch.log(probabilities[pos - 1, :]).type(torch.float)
            states["state"] += torch.tensor([[3.], [7.]])
            return logits, states

        states = {
            "state": torch.zeros(batch_size, 1)
        }

        final_ids, _ = beam_search.beam_search(
            symbols_to_logits_fn=symbols_to_logits,
            initial_ids=initial_ids,
            beam_size=beam_size,
            decode_length=decode_length,
            vocab_size=vocab_size,
            alpha=0.0,
            eos_id=1,
            states=states)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
T5 Model
"""

from typing import Optional, Union

import torch

from texar.torch.core import layers, Identity
from texar.torch.modules.embedders.embedders import WordEmbedder
from texar.torch.modules.encoder_decoders.encoder_decoder_base\
    import EncoderDecoderBase
from texar.torch.modules.encoders import T5Encoder
from texar.torch.modules.decoders import T5Decoder
from texar.torch.modules.pretrained.t5 import PretrainedT5Mixin

__all__ = [
    "T5EncoderDecoder"
]


class T5EncoderDecoder(EncoderDecoderBase, PretrainedT5Mixin):
    r"""The pre-trained T5 model. Please see
    :class:`~texar.torch.modules.PretrainedT5Mixin` for a brief description
    of T5.

    This module basically stacks
    :class:`~texar.torch.modules.WordEmbedder`,
    :class:`~texar.torch.modules.TransformerEncoder`,
    :class:`~texar.torch.modules.TransformerDecoder` and a dense
    pooler.

    Args:
        pretrained_model_name (optional): a `str`, the name
            of pre-trained model (e.g., ``T5-Small``). Please refer to
            :class:`~texar.torch.modules.PretrainedT5Mixin` for
            all supported models.
            If `None`, the model name in :attr:`hparams` is used.
        cache_dir (optional): the path to a folder in which the
            pre-trained models will be cached. If `None` (default),
            a default directory (``texar_data`` folder under user's home
            directory) will be used.
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure
            and default values.
    """

    def __init__(self,
                 pretrained_model_name: Optional[str] = None,
                 cache_dir: Optional[str] = None,
                 hparams=None):
        super().__init__(hparams=hparams)

        self.load_pretrained_config(pretrained_model_name, cache_dir)

        # Word embedding
        self.word_embedder = WordEmbedder(
            vocab_size=self._hparams.vocab_size,
            hparams=self._hparams.embed)

        # The encoder (a TransformerEncoder)
        self.encoder = T5Encoder(hparams=self._hparams.encoder)

        # The decoder (a TransformerDecoder)
        self.decoder = T5Decoder(
            token_embedder=self._embedding_fn,
            output_layer=Identity(),
            hparams=self._hparams.decoder)

        self.init_pretrained_weights()

    def _embedding_fn(self, tokens: torch.LongTensor
                      ) -> torch.Tensor:
        word_embed = self.word_embedder(tokens)
        return word_embed

    def reset_parameters(self):
        initialize = layers.get_initializer(self._hparams.initializer)
        if initialize is not None:
            # Do not re-initialize LayerNorm modules.
            for name, param in self.named_parameters():
                if name.split('.')[-1] == 'weight' and 'layer_norm' not in name:
                    initialize(param)

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.

        * The encoder arch is determined by the constructor argument
          :attr:`pretrained_model_name` if it's specified. In this case,
          `hparams` are ignored.
        * Otherwise, the encoder arch is determined by
          `hparams['pretrained_model_name']` if it's specified. All other
          configurations in `hparams` are ignored.
        * If the above two are `None`, the encoder arch is defined by the
          configurations in `hparams` and weights are randomly initialized.

        .. code-block:: python

            {
                "pretrained_model_name": "bert-base-uncased",
                "embed": {
                    "dim": 768,
                    "name": "word_embeddings"
                },
                "vocab_size": 32128,

                "encoder": {
                    "dim": 768,
                    "embedding_dropout": 0.1,
                    "multihead_attention": {
                        "dropout_rate": 0.1,
                        "name": "self",
                        "num_heads": 12,
                        "num_units": 768,
                        "output_dim": 768,
                        "use_bias": True
                    },
                    "relative_attention_num_buckets": 32,
                    "name": "t5encoder",
                    "num_blocks": 12,
                    "poswise_feedforward": {
                        "layers": [
                            {
                                "kwargs": {
                                    "in_features": 768,
                                    "out_features": 3072,
                                    "bias": True
                                },
                                "type": "Linear"
                            },
                            {"type": "ReLU"},
                            {
                                "kwargs": {
                                    "in_features": 3072,
                                    "out_features": 768,
                                    "bias": True
                                },
                                "type": "Linear"
                            }
                        ]
                    },
                    "residual_dropout": 0.1,
                    },

                "decoder": {
                    "dim": 768,
                    "embedding_dropout": 0.1,
                    "multihead_attention": {
                        "dropout_rate": 0.1,
                        "name": "self",
                        "num_heads": 12,
                        "num_units": 768,
                        "output_dim": 768,
                        "use_bias": True,
                        "relative_attention_num_buckets": 32,
                    },

                    "name": "t5coder",
                    "num_blocks": 12,
                    "poswise_feedforward": {
                        "layers": [
                            {
                                "kwargs": {
                                    "in_features": 768,
                                    "out_features": 3072,
                                    "bias": True
                                },
                                "type": "Linear"
                            },
                            {"type": "ReLU"},
                            {
                                "kwargs": {
                                    "in_features": 3072,
                                    "out_features": 768,
                                    "bias": True
                                },
                                "type": "Linear"
                            }
                        ]
                    },
                    "residual_dropout": 0.1,
                    },
                "hidden_size": 768,
                "initializer": None,
                "name": "t5_encoder_decoder",
            }

        Here:

        The default parameters are values for uncased BERT-Base model.

        `"pretrained_model_name"`: str or None
            The name of the pre-trained BERT model. If None, the model
            will be randomly initialized.

        `"embed"`: dict
            Hyperparameters for word embedding layer.

        `"vocab_size"`: int
            The vocabulary size of `inputs` in BERT model.

        `"type_vocab_size"`: int
            The vocabulary size of the `segment_ids` passed into `BertModel`.

        `"position_embed"`: dict
            Hyperparameters for position embedding layer.

        `"position_size"`: int
            The maximum sequence length that this model might ever be used with.

        `"encoder"`: dict
            Hyperparameters for the T5Encoder.
            See :func:`~texar.torch.modules.T5Encoder.default_hparams`
            for details.

        `"hidden_size"`: int
            Size of the pooler dense layer.

        `"initializer"`: dict, optional
            Hyperparameters of the default initializer that initializes
            variables created in this module.
            See :func:`~texar.torch.core.get_initializer` for details.

        `"name"`: str
            Name of the module.
        """

        return {
            'pretrained_model_name': 'T5-Small',
            'embed': {
                'dim': 768,
                'name': 'word_embeddings'
            },
            'vocab_size': 32128,

            'encoder': {
                'dim': 768,
                'embedding_dropout': 0.1,
                'multihead_attention': {
                    'dropout_rate': 0.1,
                    'name': 'self',
                    'num_heads': 12,
                    'num_units': 768,
                    'output_dim': 768,
                    'use_bias': False,
                    'is_decoder': False,
                    'relative_attention_num_buckets': 32
                },
                'eps': 1e-6,
                'name': 'encoder',
                'num_blocks': 12,
                'poswise_feedforward': {
                    'layers': [
                        {
                            'kwargs': {
                                'in_features': 768,
                                'out_features': 3072,
                                'bias': False
                            },
                            'type': 'Linear'
                        },
                        {"type": "ReLU"},
                        {
                            'kwargs': {
                                'in_features': 3072,
                                'out_features': 768,
                                'bias': False
                            },
                            'type': 'Linear'
                        }
                    ]
                },
                'residual_dropout': 0.1,
            },
            'decoder': {
                'eps': 1e-6,
                'dim': 768,
                'embedding_dropout': 0.1,
                'multihead_attention': {
                    'dropout_rate': 0.1,
                    'name': 'self',
                    'num_heads': 12,
                    'num_units': 768,
                    'output_dim': 768,
                    'use_bias': False,
                    'is_decoder': True,
                    'relative_attention_num_buckets': 32
                },
                'name': 'encoder',
                'num_blocks': 12,
                'poswise_feedforward': {
                    'layers': [
                        {
                            'kwargs': {
                                'in_features': 768,
                                'out_features': 3072,
                                'bias': False
                            },
                            'type': 'Linear'
                        },
                        {"type": "ReLU"},
                        {
                            'kwargs': {
                                'in_features': 3072,
                                'out_features': 768,
                                'bias': False
                            },
                            'type': 'Linear'
                        }
                    ]
                },
                'residual_dropout': 0.1,
            },
            'hidden_size': 768,
            'initializer': None,
            'name': 't5_encoder_decoder',
            '@no_typecheck': ['pretrained_model_name']
        }

    def forward(self,  # type: ignore
                inputs: Union[torch.Tensor, torch.LongTensor],
                sequence_length: Optional[torch.LongTensor] = None):
        r"""

        Args:
            inputs: Either a **2D Tensor** of shape `[batch_size, max_time]`,
                containing the ids of tokens in input sequences, or
                a **3D Tensor** of shape `[batch_size, max_time, vocab_size]`,
                containing soft token ids (i.e., weights or probabilities)
                used to mix the embedding vectors.
            sequence_length: A 1D :tensor:`Tensor` of shape
                ``[batch_size]``. Input tokens beyond respective sequence
                lengths are masked out automatically.

        Returns:
        """
        if inputs.dim() == 2:
            word_embeds = self.word_embedder(ids=inputs)
        elif inputs.dim() == 3:
            word_embeds = self.word_embedder(soft_ids=inputs)
        else:
            raise ValueError("'inputs' should be a 2D or 3D tensor.")

        batch_size = inputs.size(0)

        if sequence_length is None:
            sequence_length = inputs.new_full(
                (batch_size,), inputs.size(1), dtype=torch.long)

        encoder_output = self.encoder(inputs=word_embeds,
                                      sequence_length=sequence_length)

        decoder_output = self.decoder(inputs=inputs,
                                      memory=encoder_output,
                                      memory_sequence_length=sequence_length)

        return encoder_output, decoder_output

    @property
    def output_size(self):
        r"""The feature size of :meth:`forward` output
        :attr:`pooled_output`.
        """
        return self._hparams.hidden_size



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Various helper utilities for attention mechanism.

The implementation of `sparsemax` adapted from:
    `https://github.com/OpenNMT/OpenNMT-py/blob/master/onmt/
    modules/sparse_activations.py`
"""

from typing import Optional, Tuple

import torch
import torch.nn.functional as F
from torch.autograd import Function

from texar.torch.utils.utils import sequence_mask

__all__ = [
    'hardmax',
    'maybe_mask_score',
    'prepare_memory',
    'safe_cumprod',
    'sparsemax',
]


def hardmax(logits: torch.Tensor) -> torch.Tensor:
    r"""Returns batched one-hot vectors. The depth index containing
    the `1` is that of the maximum logit value.

    Args:
        logits: A batch tensor of logit values.

    Returns:
        A batched one-hot tensor.
    """
    if not isinstance(logits, torch.Tensor):
        logits = torch.tensor(logits)
    depth = logits.shape[-1]
    one_hot = torch.eye(depth, dtype=torch.int64)
    return F.embedding(torch.argmax(logits, -1), one_hot)


def maybe_mask_score(score: torch.Tensor,
                     score_mask_value: torch.Tensor,
                     memory_sequence_length: Optional[torch.LongTensor]) \
        -> torch.Tensor:
    r"""Mask the attention score based on the masks."""
    if memory_sequence_length is None:
        return score

    for memory_sequence_length_value in memory_sequence_length:
        if memory_sequence_length_value <= 0:
            raise ValueError(
                "All values in memory_sequence_length must be greater "
                "than zero.")

    score_mask = sequence_mask(memory_sequence_length,
                               max_len=score.shape[1])
    score_mask_values = score_mask_value * torch.ones_like(score)
    return torch.where(score_mask, score, score_mask_values)


def prepare_memory(memory: torch.Tensor,
                   memory_sequence_length: Optional[torch.LongTensor]) \
        -> torch.Tensor:
    r"""Convert to tensor and possibly mask ``memory``.

    Args:
        memory: tensor, shaped ``[batch_size, max_time, ...]``.
        memory_sequence_length: integer tensor, shaped ``[batch_size]``.

    Returns:
        A (possibly masked), new ``memory``.

    Raises:
        ValueError: if ``memory`` and ``memory_sequence_length`` do not have
        the same ``batch_size``.
    """
    if (memory_sequence_length is not None and
            not isinstance(memory_sequence_length, torch.Tensor)):
        memory_sequence_length = torch.tensor(memory_sequence_length,
                                              dtype=torch.long,
                                              device=memory.device)

    if memory_sequence_length is None:
        seq_len_mask = None
    else:
        seq_len_mask = sequence_mask(memory_sequence_length,
                                     max_len=memory.shape[1],
                                     dtype=memory.dtype)
        seq_len_batch_size = memory_sequence_length.shape[0]

    # Mask the memory based on the memory mask.
    rank = memory.dim()
    m_batch_size = memory.shape[0]

    if seq_len_mask is not None:
        if seq_len_batch_size != m_batch_size:
            raise ValueError("memory_sequence_length and memory tensor "
                             "batch sizes do not match.")
        return memory * seq_len_mask.view(
            seq_len_mask.size() + (1,) * (rank - 2))
    else:
        return memory


def safe_cumprod(x: torch.Tensor,
                 *args,
                 **kwargs) -> torch.Tensor:
    r"""Computes cumprod of x in logspace using cumsum to avoid underflow.
    The cumprod function and its gradient can result in numerical
    instabilities when its argument has very small and/or zero values.
    As long as the argument is all positive, we can instead compute the
    cumulative product as `exp(cumsum(log(x)))`.  This function can be called
    identically to :torch:`cumprod`.

    Args:
        x: Tensor to take the cumulative product of.
        *args: Passed on to cumsum; these are identical to those in cumprod.
        **kwargs: Passed on to cumsum; these are identical to those in cumprod.

    Returns:
        Cumulative product of x.
    """
    if not isinstance(x, torch.Tensor):
        x = torch.tensor(x)

    tiny = torch.finfo(x.dtype).tiny

    return torch.exp(torch.cumsum(torch.log(torch.clamp(x, tiny, 1)),
                                  *args, **kwargs))


def _make_ix_like(input: torch.Tensor,
                  dim: int = -1) -> torch.Tensor:
    d = input.size(dim)
    rho = torch.arange(1, d + 1, device=input.device, dtype=input.dtype)
    view = (-1,) + (1,) * (input.dim() - 1)
    return rho.view(view).transpose(0, dim)


def _threshold_and_support(input: torch.Tensor,
                           dim: int = -1) -> Tuple[torch.Tensor, torch.Tensor]:
    r"""`Sparsemax` building block: compute the threshold.

    Args:
        input: any dimension
        dim: dimension along which to apply `sparsemax`.

    Returns:
        the threshold value
    """

    input_srt, _ = torch.sort(input, descending=True, dim=dim)
    input_cumsum = input_srt.cumsum(dim) - 1
    rhos = _make_ix_like(input, dim)
    support = rhos * input_srt > input_cumsum

    support_size = support.sum(dim=dim).unsqueeze(dim)
    tau = input_cumsum.gather(dim, support_size - 1)
    tau /= support_size.to(input.dtype)
    return tau, support_size


class SparsemaxFunction(Function):

    @staticmethod
    def forward(ctx,  # type: ignore
                input: torch.Tensor,
                dim: int = -1) -> torch.Tensor:
        ctx.dim = dim
        max_val, _ = input.max(dim=dim, keepdim=True)
        input -= max_val  # same numerical stability trick as for softmax
        tau, supp_size = _threshold_and_support(input, dim=dim)
        output = torch.clamp(input - tau, min=0)
        ctx.save_for_backward(supp_size, output)
        return output

    @staticmethod
    def backward(ctx,  # type: ignore
                 grad_output: torch.Tensor) -> Tuple[torch.Tensor, None]:
        supp_size, output = ctx.saved_tensors
        dim = ctx.dim
        grad_input = grad_output.clone()
        grad_input[output == 0] = 0

        v_hat = grad_input.sum(dim=dim) / supp_size.to(output.dtype).squeeze()
        v_hat = v_hat.unsqueeze(dim)
        grad_input = torch.where(output != 0, grad_input - v_hat, grad_input)
        return grad_input, None


def sparsemax(input: torch.Tensor, dim: int = -1) -> torch.Tensor:
    r"""`sparsemax`: normalizing sparse transform (a la softmax).

    Args:
        input (Tensor): A batch tensor of logit values.
        dim: Dimension along which to apply `sparsemax`.

    Returns:
        Tensor: output with the same shape as input.
    """
    return SparsemaxFunction.apply(input, dim)  # type: ignore



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for RNN decoders.
"""

import unittest

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from texar.torch.hyperparams import HParams
from texar.torch.modules.decoders.decoder_helpers import get_helper
from texar.torch.modules.decoders.rnn_decoders import (
    AttentionRNNDecoder, AttentionRNNDecoderOutput, BasicRNNDecoder,
    BasicRNNDecoderOutput)
from texar.torch.modules.embedders.embedders import WordEmbedder
from texar.torch.utils.utils import map_structure


class BasicRNNDecoderTest(unittest.TestCase):
    r"""Tests :class:`~texar.torch.modules.decoders.rnn_decoders.BasicRNNDecoder`.
    """

    def setUp(self):
        self._vocab_size = 4
        self._max_time = 8
        self._batch_size = 16
        self._emb_dim = 20
        self._inputs = torch.randint(
            self._vocab_size, size=(self._batch_size, self._max_time))
        embedding = torch.rand(
            self._vocab_size, self._emb_dim, dtype=torch.float)
        self._embedder = WordEmbedder(init_value=embedding)
        self._hparams = HParams(None, BasicRNNDecoder.default_hparams())

    def _test_outputs(self, decoder, outputs, final_state, sequence_lengths,
                      test_mode=False):
        hidden_size = decoder.hparams.rnn_cell.kwargs.num_units

        self.assertIsInstance(outputs, BasicRNNDecoderOutput)
        max_time = (self._max_time if not test_mode
                    else max(sequence_lengths).item())
        self.assertEqual(
            outputs.logits.shape,
            (self._batch_size, max_time, self._vocab_size))
        if not test_mode:
            np.testing.assert_array_equal(
                sequence_lengths, [max_time] * self._batch_size)
        self.assertEqual(final_state[0].shape, (self._batch_size, hidden_size))

    def test_decode_train(self):
        r"""Tests decoding in training mode.
        """
        decoder = BasicRNNDecoder(
            token_embedder=self._embedder, input_size=self._emb_dim,
            vocab_size=self._vocab_size, hparams=self._hparams)
        sequence_length = torch.tensor([self._max_time] * self._batch_size)

        # Helper by default HParams
        helper_train = decoder.create_helper()
        outputs, final_state, sequence_lengths = decoder(
            helper=helper_train, inputs=self._inputs,
            sequence_length=sequence_length)
        self._test_outputs(decoder, outputs, final_state, sequence_lengths)

        # Helper by decoding strategy
        helper_train = decoder.create_helper(decoding_strategy='train_greedy')
        outputs, final_state, sequence_lengths = decoder(
            helper=helper_train, inputs=self._inputs,
            sequence_length=sequence_length)
        self._test_outputs(decoder, outputs, final_state, sequence_lengths)

        # Implicit helper
        outputs, final_state, sequence_lengths = decoder(
            inputs=self._inputs, sequence_length=sequence_length)
        self._test_outputs(decoder, outputs, final_state, sequence_lengths)

        # Eval helper through forward args
        outputs, final_state, sequence_lengths = decoder(
            embedding=self._embedder,
            start_tokens=torch.tensor([1] * self._batch_size),
            end_token=2, infer_mode=True)
        self._test_outputs(
            decoder, outputs, final_state, sequence_lengths, test_mode=True)

    @staticmethod
    def _assert_tensor_equal(a: torch.Tensor, b: torch.Tensor) -> bool:
        if torch.is_tensor(a):
            a = a.detach().numpy()
        if torch.is_tensor(b):
            b = b.detach().numpy()
        if any(np.issubdtype(array.dtype, np.floating) for array in [a, b]):
            return np.testing.assert_allclose(a, b, rtol=1e-5, atol=1e-8)
        return np.testing.assert_array_equal(a, b)

    def test_decode_train_with_torch(self):
        r"""Compares decoding results with PyTorch built-in decoder.
        """
        decoder = BasicRNNDecoder(
            token_embedder=self._embedder, input_size=self._emb_dim,
            vocab_size=self._vocab_size, hparams=self._hparams)

        input_size = self._emb_dim
        hidden_size = decoder.hparams.rnn_cell.kwargs.num_units
        num_layers = decoder.hparams.rnn_cell.num_layers
        torch_lstm = nn.LSTM(input_size, hidden_size, num_layers,
                             batch_first=True)

        # match parameters
        for name in ['weight_ih', 'weight_hh', 'bias_ih', 'bias_hh']:
            setattr(torch_lstm, f'{name}_l0',
                    getattr(decoder._cell._cell, name))
        torch_lstm.flatten_parameters()

        output_layer = decoder._output_layer
        input_lengths = torch.tensor([self._max_time] * self._batch_size)
        inputs = torch.randint(
            self._vocab_size, size=(self._batch_size, self._max_time))

        # decoder outputs
        helper_train = decoder.create_helper()
        outputs, final_state, sequence_lengths = decoder(
            inputs=inputs,
            sequence_length=input_lengths,
            helper=helper_train)

        # torch LSTM outputs
        lstm_inputs = F.embedding(inputs, self._embedder.embedding)
        torch_outputs, torch_states = torch_lstm(lstm_inputs)
        torch_outputs = output_layer(torch_outputs)
        torch_sample_id = torch.argmax(torch_outputs, dim=-1)

        self.assertEqual(final_state[0].shape,
                         (self._batch_size, hidden_size))

        self._assert_tensor_equal(outputs.logits, torch_outputs)
        self._assert_tensor_equal(outputs.sample_id, torch_sample_id)
        self._assert_tensor_equal(final_state[0], torch_states[0].squeeze(0))
        self._assert_tensor_equal(final_state[1], torch_states[1].squeeze(0))
        self._assert_tensor_equal(sequence_lengths, input_lengths)

    def test_decode_infer(self):
        r"""Tests decoding in inference mode."""
        decoder = BasicRNNDecoder(
            token_embedder=self._embedder, input_size=self._emb_dim,
            vocab_size=self._vocab_size, hparams=self._hparams)

        decoder.eval()
        start_tokens = torch.tensor([self._vocab_size - 2] * self._batch_size)

        helpers = []
        for strategy in ['infer_greedy', 'infer_sample']:
            helper = decoder.create_helper(
                decoding_strategy=strategy,
                start_tokens=start_tokens,
                end_token=self._vocab_size - 1)
            helpers.append(helper)
        for klass in ['TopKSampleEmbeddingHelper', 'SoftmaxEmbeddingHelper',
                      'GumbelSoftmaxEmbeddingHelper']:
            helper = get_helper(
                klass, start_tokens=start_tokens,
                end_token=self._vocab_size - 1,
                top_k=self._vocab_size // 2, tau=2.0,
                straight_through=True)
            helpers.append(helper)

        for helper in helpers:
            max_length = 100
            outputs, final_state, sequence_lengths = decoder(
                helper=helper, max_decoding_length=max_length)
            self.assertLessEqual(max(sequence_lengths), max_length)
            self._test_outputs(decoder, outputs, final_state, sequence_lengths,
                               test_mode=True)


class AttentionRNNDecoderTest(unittest.TestCase):
    r"""Tests :class:`~texar.torch.modules.decoders.rnn_decoders.AttentionRNNDecoder`.
    """

    def setUp(self):
        self._vocab_size = 10
        self._max_time = 16
        self._batch_size = 8
        self._emb_dim = 20
        self._attention_dim = 256
        self._inputs = torch.randint(
            self._vocab_size, size=(self._batch_size, self._max_time))
        embedding = torch.rand(
            self._vocab_size, self._emb_dim, dtype=torch.float)
        self._embedder = WordEmbedder(init_value=embedding)
        self._encoder_output = torch.rand(
            self._batch_size, self._max_time, 64)

        self._test_hparams = {}  # (cell_type, is_multi) -> hparams
        for cell_type in ["RNNCell", "LSTMCell", "GRUCell"]:
            hparams = {
                "rnn_cell": {
                    'type': cell_type,
                    'kwargs': {
                        'num_units': 256,
                    },
                },
                "attention": {
                    "kwargs": {
                        "num_units": self._attention_dim
                    },
                }
            }
            self._test_hparams[(cell_type, False)] = HParams(
                hparams, AttentionRNNDecoder.default_hparams())

        hparams = {
            "rnn_cell": {
                'type': 'LSTMCell',
                'kwargs': {
                    'num_units': 256,
                },
                'num_layers': 3,
            },
            "attention": {
                "kwargs": {
                    "num_units": self._attention_dim
                },
            }
        }
        self._test_hparams[("LSTMCell", True)] = HParams(
            hparams, AttentionRNNDecoder.default_hparams())

    def _test_outputs(self, decoder, outputs, final_state, sequence_lengths,
                      test_mode=False):
        hidden_size = decoder.hparams.rnn_cell.kwargs.num_units
        cell_type = decoder.hparams.rnn_cell.type
        is_multi = decoder.hparams.rnn_cell.num_layers > 1

        self.assertIsInstance(outputs, AttentionRNNDecoderOutput)
        max_time = (self._max_time if not test_mode
                    else max(sequence_lengths).item())
        self.assertEqual(
            outputs.logits.shape,
            (self._batch_size, max_time, self._vocab_size))
        if not test_mode:
            np.testing.assert_array_equal(
                sequence_lengths, [max_time] * self._batch_size)

        map_structure(
            lambda t: self.assertEqual(
                t.size(), (self._batch_size, hidden_size)),
            final_state.cell_state)
        state = final_state.cell_state
        if is_multi:
            self.assertIsInstance(state, list)
            state = state[0]
        if cell_type == "LSTMCell":
            self.assertIsInstance(state, tuple)
            state = state[0]
        self.assertIsInstance(state, torch.Tensor)

    def test_decode_infer(self):
        r"""Tests decoding in inference mode.
        """
        seq_length = np.random.randint(
            self._max_time, size=[self._batch_size]) + 1
        encoder_values_length = torch.tensor(seq_length)

        for (cell_type, is_multi), hparams in self._test_hparams.items():
            decoder = AttentionRNNDecoder(
                encoder_output_size=64,
                token_embedder=self._embedder,
                vocab_size=self._vocab_size,
                input_size=self._emb_dim,
                hparams=hparams)

            decoder.eval()

            helper_infer = decoder.create_helper(
                start_tokens=torch.tensor([1] * self._batch_size), end_token=2)

            outputs, final_state, sequence_lengths = decoder(
                memory=self._encoder_output,
                memory_sequence_length=encoder_values_length,
                helper=helper_infer)

            self._test_outputs(decoder, outputs, final_state, sequence_lengths,
                               test_mode=True)

    def test_decode_train(self):
        r"""Tests decoding in training mode.
        """
        seq_length = np.random.randint(
            self._max_time, size=[self._batch_size]) + 1
        encoder_values_length = torch.tensor(seq_length)

        for (cell_type, is_multi), hparams in self._test_hparams.items():
            decoder = AttentionRNNDecoder(
                encoder_output_size=64,
                token_embedder=self._embedder,
                vocab_size=self._vocab_size,
                input_size=self._emb_dim,
                hparams=hparams)

            sequence_length = torch.tensor([self._max_time] * self._batch_size)

            helper_train = decoder.create_helper()
            outputs, final_state, sequence_lengths = decoder(
                memory=self._encoder_output,
                memory_sequence_length=encoder_values_length,
                helper=helper_train,
                inputs=self._inputs,
                sequence_length=sequence_length)

            self._test_outputs(decoder, outputs, final_state, sequence_lengths)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for sampler related operations.
"""
import unittest
from typing import no_type_check

import numpy as np

from texar.torch.data.data.data_base import (
    DatasetBase, DataSource, IterDataSource, SequenceDataSource)
from texar.torch.data.data.sampler import BufferShuffleSampler


class SamplerTest(unittest.TestCase):
    r"""Tests samplers.
    """

    class MockDataBase(DatasetBase):
        def __init__(self, size: int, lazy_strategy: str,
                     cache_strategy: str, unknown_size: bool = False):
            data = list(range(size))
            source: DataSource[int]
            if unknown_size:
                source = IterDataSource(data)
            else:
                source = SequenceDataSource(data)
            hparams = {
                'lazy_strategy': lazy_strategy,
                'cache_strategy': cache_strategy,
            }
            super().__init__(source, hparams=hparams)

    def setUp(self) -> None:
        self.size = 10
        self.buffer_size = 5

    @no_type_check
    def _test_data(self, data: DatasetBase,
                   returns_data: bool = False,
                   always_returns_data: bool = False):
        sampler = BufferShuffleSampler(data, self.buffer_size)
        for epoch in range(2):
            indices = list(iter(sampler))
            if always_returns_data or (returns_data and epoch == 0):
                examples = [ex[1] for ex in indices]
                indices = [ex[0] for ex in indices]
                np.testing.assert_array_equal(indices, examples)
            self.assertEqual(len(set(indices)), self.size)
            self.assertEqual(min(indices), 0)
            self.assertEqual(max(indices), self.size - 1)
            data._fully_cached = True

    def test_known_size(self):
        data = self.MockDataBase(self.size, 'none', 'processed')
        self._test_data(data)
        data = self.MockDataBase(self.size, 'all', 'none', unknown_size=True)
        self._test_data(data, always_returns_data=True)

    def test_non_lazy_loading(self):
        strategies = [
            ('none', 'processed'),
            ('process', 'loaded'),
            ('process', 'processed'),
        ]
        for lazy, cache in strategies:
            data = self.MockDataBase(self.size, lazy, cache)
            self._test_data(data)

    def test_lazy_loading(self):
        data = self.MockDataBase(self.size, 'all', 'loaded', unknown_size=True)
        self._test_data(data, returns_data=True)
        data = self.MockDataBase(self.size, 'all', 'processed',
                                 unknown_size=True)
        self._test_data(data, returns_data=True)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for regularizers.
"""

import unittest

import torch

from texar.torch.core.regularizers import *


class RegularizerTest(unittest.TestCase):
    r"""Test regularizers.
    """

    def setUp(self):
        self.x = torch.tensor([-1, 2, -3, 4, -5])
        self.l1 = 0.1
        self.l2 = 0.2

    def test_l1(self):
        r"""Tests l1."""
        regularizer = l1(self.l1)
        self.assertEqual(regularizer.l1, 0.1)
        self.assertEqual(regularizer.l2, 0)

        regularization = regularizer(self.x)
        self.assertEqual(regularization.item(), self.l1 * (1 + 2 + 3 + 4 + 5))

        self.assertEqual(regularizer.get_config().get("l1"), 0.1)
        self.assertEqual(regularizer.get_config().get("l2"), 0)

    def test_l2(self):
        r"""Tests l2."""
        regularizer = l2(self.l2)
        self.assertEqual(regularizer.l1, 0)
        self.assertEqual(regularizer.l2, 0.2)

        regularization = regularizer(self.x)
        self.assertEqual(regularization.item(), self.l2 * (1 + 4 + 9 + 16 + 25))

        self.assertEqual(regularizer.get_config().get("l1"), 0)
        self.assertEqual(regularizer.get_config().get("l2"), 0.2)

    def test_l1_l2(self):
        r"""Tests l1_l2."""
        regularizer = l1_l2(self.l1, self.l2)
        self.assertEqual(regularizer.l1, 0.1)
        self.assertEqual(regularizer.l2, 0.2)

        regularization = regularizer(self.x)
        self.assertEqual(regularization.item(), self.l1 * (1 + 2 + 3 + 4 + 5) +
                         self.l2 * (1 + 4 + 9 + 16 + 25))

        self.assertEqual(regularizer.get_config().get("l1"), 0.1)
        self.assertEqual(regularizer.get_config().get("l2"), 0.2)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various helper classes and utilities for RNN decoders.
"""
from abc import ABC
from typing import Callable, Generic, Optional, Tuple, Type, TypeVar, Union

import torch
import torch.nn.functional as F
from torch.distributions import Categorical, Gumbel

from texar.torch.utils import utils
from texar.torch.utils.dtypes import torch_bool

__all__ = [
    'Helper',
    'TrainingHelper',
    'EmbeddingHelper',
    'GreedyEmbeddingHelper',
    'SampleEmbeddingHelper',
    'TopKSampleEmbeddingHelper',
    'TopPSampleEmbeddingHelper',
    'SoftmaxEmbeddingHelper',
    'GumbelSoftmaxEmbeddingHelper',
    'default_helper_train_hparams',
    'default_helper_infer_hparams',
    'get_helper',
]
# TODO: Implement `ScheduledEmbeddingTrainingHelper` and
#     `ScheduledOutputTrainingHelper`

HelperInitTuple = Tuple[torch.ByteTensor, torch.Tensor]
NextInputTuple = Tuple[torch.ByteTensor, torch.Tensor]

# indices, position -> embeddings
EmbeddingFn = Callable[[torch.LongTensor, torch.LongTensor], torch.Tensor]

IDType = TypeVar('IDType', bound=torch.Tensor)


# Helper instances are used by :class:`texar.torch.modules.DecoderBase`.
class Helper(Generic[IDType], ABC):
    r"""Interface for implementing sampling in seq2seq decoders.

    Please refer to the documentation for the TensorFlow counterpart
    `tf.contrib.seq2seq.Helper
    <https://www.tensorflow.org/api_docs/python/tf/contrib/seq2seq/Helper>`_.
    """

    def initialize(self, embedding_fn: EmbeddingFn,
                   inputs: Optional[torch.Tensor],
                   sequence_length: Optional[torch.LongTensor]) \
            -> HelperInitTuple:
        r"""Initialize the current batch.

        Args:
            embedding_fn: A function taking input tokens and timestamps,
                returning embedding tensors.
            inputs: Input tensors.
            sequence_length: An int32 vector tensor.

        Returns:
            ``(initial_finished, initial_inputs)``.
        """
        raise NotImplementedError

    def sample(self, time: int, outputs: torch.Tensor) -> IDType:
        r"""Returns ``sample_ids``.
        """
        raise NotImplementedError

    def next_inputs(self, embedding_fn: EmbeddingFn,
                    time: int, outputs: torch.Tensor,
                    sample_ids: IDType) -> NextInputTuple:
        r"""Returns ``(finished, next_inputs, next_state)``.
        """
        raise NotImplementedError


class TrainingHelper(Helper[torch.LongTensor]):
    r"""A helper for use during training. Only reads inputs.

    Returned ``sample_ids`` are the argmax of the RNN output logits.

    Please refer to the documentation for the TensorFlow counterpart
    `tf.contrib.seq2seq.TrainingHelper
    <https://www.tensorflow.org/api_docs/python/tf/contrib/seq2seq/TrainingHelper>`_.

    Args:
        time_major (bool):  Whether the tensors in ``inputs`` are time major.
            If `False` (default), they are assumed to be batch major.
    """
    # the following are set in `initialize`
    _inputs: torch.Tensor
    _zero_inputs: torch.Tensor
    _sequence_length: torch.LongTensor

    def __init__(self, time_major: bool = False):
        self._time_major = time_major

    def initialize(self, embedding_fn: EmbeddingFn,
                   inputs: Optional[torch.Tensor],
                   sequence_length: Optional[torch.LongTensor]) \
            -> HelperInitTuple:
        if inputs is None:
            raise ValueError("`inputs` cannot be None for TrainingHelper")
        if sequence_length is None:
            raise ValueError(
                "`sequence_length` cannot be None for TrainingHelper")
        inputs: torch.Tensor
        sequence_length: torch.LongTensor

        if sequence_length.dim() != 1:
            raise ValueError(
                f"Expected 'sequence_length' to be a vector, "
                f"but received shape: {sequence_length.shape}")

        if not self._time_major:
            inputs = inputs.transpose(0, 1)  # make inputs time major
        times = torch.arange(
            sequence_length.max(), dtype=torch.long, device=inputs.device)
        times = times.unsqueeze(1).expand(-1, inputs.size(1))
        inputs = embedding_fn(inputs, times)

        self._inputs = inputs
        self._sequence_length = sequence_length
        self._zero_inputs = inputs.new_zeros(inputs[0].size())

        finished: torch.ByteTensor = (sequence_length == 0)
        all_finished = torch.all(finished).item()
        next_inputs = inputs[0] if not all_finished else self._zero_inputs
        return (finished, next_inputs)

    def sample(self, time: int, outputs: torch.Tensor) -> torch.LongTensor:
        del time
        sample_ids = torch.argmax(outputs, dim=-1)
        return sample_ids

    def next_inputs(self, embedding_fn: EmbeddingFn,
                    time: int, outputs: torch.Tensor,
                    sample_ids: torch.LongTensor) -> NextInputTuple:
        del embedding_fn, outputs, sample_ids
        next_time = time + 1
        finished = (next_time >= self._sequence_length)
        all_finished = torch.all(finished).item()

        next_inputs = (self._inputs[next_time] if not all_finished
                       else self._zero_inputs)
        return (finished, next_inputs)


class EmbeddingHelper(Helper[IDType], ABC):
    r"""A generic helper for use during inference.

    Uses output logits for sampling, and passes the result through an embedding
    layer to get the next input.

    Args:
        start_tokens: 1D :tensor:`LongTensor` shaped ``[batch_size]``,
            representing the start tokens for each sequence in batch.
        end_token: Python int or scalar :tensor:`LongTensor`, denoting the
            token that marks end of decoding.

    Raises:
        ValueError: if :attr:`start_tokens` is not a 1D tensor or
            :attr:`end_token` is not a scalar.
    """

    _start_inputs: torch.Tensor  # set in `initialize`

    def __init__(self, start_tokens: torch.LongTensor,
                 end_token: Union[int, torch.LongTensor]):
        if start_tokens.dim() != 1:
            raise ValueError("start_tokens must be a vector")
        if not isinstance(end_token, int) and end_token.dim() != 0:
            raise ValueError("end_token must be a scalar")

        self._start_tokens = start_tokens
        self._batch_size = start_tokens.size(0)
        if isinstance(end_token, int):
            self._end_token = start_tokens.new_tensor(end_token)
        else:
            self._end_token = end_token

    @property
    def batch_size(self) -> int:
        return self._batch_size

    def initialize(self, embedding_fn: EmbeddingFn,
                   inputs: Optional[torch.Tensor],
                   sequence_length: Optional[torch.LongTensor]) \
            -> HelperInitTuple:
        del inputs, sequence_length
        times = torch.zeros_like(self._start_tokens)
        self._start_inputs = embedding_fn(self._start_tokens, times)
        finished = torch.zeros_like(self._start_tokens, dtype=torch_bool)
        return (finished, self._start_inputs)


class SingleEmbeddingHelper(EmbeddingHelper[torch.LongTensor], ABC):
    r"""A generic helper for use during inference.

    Based on :class:`~texar.torch.modules.EmbeddingHelper`, this class returns
    samples that are vocabulary indices, and use the corresponding embeddings as
    the next input. This class also supports callables as embeddings.

    Args:
        start_tokens: 1D :tensor:`LongTensor` shaped ``[batch_size]``,
            representing the start tokens for each sequence in batch.
        end_token: Python int or scalar :tensor:`LongTensor`, denoting the
            token that marks end of decoding.

    Raises:
        ValueError: if :attr:`start_tokens` is not a 1D tensor or
            :attr:`end_token` is not a scalar.
    """

    def next_inputs(self, embedding_fn: EmbeddingFn,
                    time: int, outputs: torch.Tensor,
                    sample_ids: torch.LongTensor) -> NextInputTuple:
        del outputs  # unused by next_inputs_fn
        finished = (sample_ids == self._end_token)
        all_finished = torch.all(finished).item()

        times = torch.full_like(sample_ids, time + 1)
        embeddings = embedding_fn(sample_ids, times)

        next_inputs = (embeddings if not all_finished else self._start_inputs)
        return (finished, next_inputs)


class GreedyEmbeddingHelper(SingleEmbeddingHelper):
    r"""A helper for use during inference.

    Uses the argmax of the output (treated as logits) and passes the
    result through an embedding layer to get the next input.

    Note that for greedy decoding, Texar's decoders provide a simpler
    interface by specifying ``decoding_strategy='infer_greedy'`` when calling a
    decoder (see, e.g.,,
    :meth:`RNN decoder <texar.torch.modules.RNNDecoderBase.forward>`). In this
    case, use of :class:`GreedyEmbeddingHelper` is not necessary.

    Please refer to the documentation for the TensorFlow counterpart
    `tf.contrib.seq2seq.GreedyEmbeddingHelper
    <https://www.tensorflow.org/api_docs/python/tf/contrib/seq2seq/GreedyEmbeddingHelper>`_.

    Args:
        start_tokens: 1D :tensor:`LongTensor` shaped ``[batch_size]``,
            representing the start tokens for each sequence in batch.
        end_token: Python int or scalar :tensor:`LongTensor`, denoting the
            token that marks end of decoding.

    Raises:
        ValueError: if :attr:`start_tokens` is not a 1D tensor or
            :attr:`end_token` is not a scalar.
    """

    def sample(self, time: int, outputs: torch.Tensor) -> torch.LongTensor:
        del time  # unused by sample_fn
        # Outputs are logits, use argmax to get the most probable id
        if not torch.is_tensor(outputs):
            raise TypeError(
                f"Expected outputs to be a single Tensor, got: {type(outputs)}")
        sample_ids = torch.argmax(outputs, dim=-1)
        return sample_ids


class SampleEmbeddingHelper(SingleEmbeddingHelper):
    r"""A helper for use during inference.

    Uses sampling (from a distribution) instead of argmax and passes the
    result through an embedding layer to get the next input.

    Please refer to the documentation for the TensorFlow counterpart
    `tf.contrib.seq2seq.SampleEmbeddingHelper
    <https://www.tensorflow.org/api_docs/python/tf/contrib/seq2seq/SampleEmbeddingHelper>`_.

    Args:
        embedding: A callable or the ``params`` argument for
            :torch_nn:`functional.embedding`.
            If a callable, it can take a vector tensor of ``ids`` (argmax
            ids), or take two arguments (``ids``, ``times``), where ``ids``
            is a vector of argmax ids, and ``times`` is a vector of current
            time steps (i.e., position ids). The latter case can be used
            when :attr:`embedding` is a combination of word embedding and
            position embedding.
            The returned tensor will be passed to the decoder input.
        start_tokens: 1D :tensor:`LongTensor` shaped ``[batch_size]``,
            representing the start tokens for each sequence in batch.
        end_token: Python int or scalar :tensor:`LongTensor`, denoting the
            token that marks end of decoding.
        softmax_temperature (float, optional): Value to divide the logits by
            before computing the softmax. Larger values (above 1.0) result
            in more random samples, while smaller values push the sampling
            distribution towards the argmax. Must be strictly greater than
            0. Defaults to 1.0.

    Raises:
        ValueError: if :attr:`start_tokens` is not a 1D tensor or
            :attr:`end_token` is not a scalar.
    """

    def __init__(self, start_tokens: torch.LongTensor,
                 end_token: Union[int, torch.LongTensor],
                 softmax_temperature: Optional[float] = None):
        super().__init__(start_tokens, end_token)
        self._softmax_temperature = softmax_temperature

    def sample(self, time: int, outputs: torch.Tensor) -> torch.LongTensor:
        del time  # unused by sample_fn
        # Outputs are logits, we sample instead of argmax (greedy).
        if not torch.is_tensor(outputs):
            raise TypeError(
                f"Expected outputs to be a single Tensor, got: {type(outputs)}")
        if self._softmax_temperature is None:
            logits = outputs
        else:
            logits = outputs / self._softmax_temperature

        sample_id_sampler = Categorical(logits=logits)
        sample_ids = sample_id_sampler.sample()

        return sample_ids


def _top_k_logits(logits: torch.Tensor, k: int) -> torch.Tensor:
    r"""Adapted from
    https://github.com/openai/gpt-2/blob/master/src/sample.py#L63-L77
    """
    if k == 0:
        # no truncation
        return logits

    values, _ = torch.topk(logits, k=k)
    min_values: torch.Tensor = values[:, -1].unsqueeze(-1)
    return torch.where(
        logits < min_values,
        torch.full_like(logits, float('-inf')), logits)


def _top_p_logits(logits: torch.Tensor, p: float) -> torch.Tensor:
    r"""Adapted from
    https://gist.github.com/thomwolf/1a5a29f6962089e871b94cbd09daf317#file-top-k-top-p-py-L16-L27"""
    sorted_logits, sorted_indices = torch.sort(logits, descending=True)
    cumulative_probs = torch.cumsum(F.softmax(sorted_logits, dim=-1), dim=-1)

    # Remove tokens with cumulative probability above the threshold
    sorted_indices_to_remove = cumulative_probs > p
    # Shift the indices to the right to keep also the first token above the
    # threshold
    sorted_indices_to_remove[:, 1:] = sorted_indices_to_remove[:, :-1].clone()
    sorted_indices_to_remove[:, 0] = 0

    for idx in range(logits.size(0)):
        batch_indices = sorted_indices[idx, sorted_indices_to_remove[idx]]
        logits[idx, batch_indices] = float("-inf")
    return logits


class TopKSampleEmbeddingHelper(SingleEmbeddingHelper):
    r"""A helper for use during inference.

    Samples from ``top_k`` most likely candidates from a vocab distribution,
    and passes the result through an embedding layer to get the next input.

    Args:
        start_tokens: 1D :tensor:`LongTensor` shaped ``[batch_size]``,
            representing the start tokens for each sequence in batch.
        end_token: Python int or scalar :tensor:`LongTensor`, denoting the
            token that marks end of decoding.
        top_k (int, optional): Number of top candidates to sample from. Must
            be `>=0`. If set to 0, samples from all candidates (i.e.,
            regular random sample decoding). Defaults to 10.
        softmax_temperature (float, optional): Value to divide the logits by
            before computing the softmax. Larger values (above 1.0) result
            in more random samples, while smaller values push the sampling
            distribution towards the argmax. Must be strictly greater than
            0. Defaults to 1.0.

    Raises:
        ValueError: if :attr:`start_tokens` is not a 1D tensor or
            :attr:`end_token` is not a scalar.
    """

    def __init__(self, start_tokens: torch.LongTensor,
                 end_token: Union[int, torch.LongTensor], top_k: int = 10,
                 softmax_temperature: Optional[float] = None):
        super().__init__(start_tokens, end_token)
        self._top_k = top_k
        self._softmax_temperature = softmax_temperature

    def sample(self, time: int, outputs: torch.Tensor) -> torch.LongTensor:
        del time  # unused by sample_fn
        # Outputs are logits, we sample from the top-k candidates
        if not torch.is_tensor(outputs):
            raise TypeError(
                f"Expected outputs to be a single Tensor, got: {type(outputs)}")
        if self._softmax_temperature is None:
            logits = outputs
        else:
            logits = outputs / self._softmax_temperature

        logits = _top_k_logits(logits, k=self._top_k)

        sample_id_sampler = Categorical(logits=logits)
        sample_ids = sample_id_sampler.sample()

        return sample_ids


class TopPSampleEmbeddingHelper(SingleEmbeddingHelper):
    r"""A helper for use during inference.

    Samples from candidates that have a cumulative probability of at most `p`
    when arranged in decreasing order, and passes the result through an
    embedding layer to get the next input. This is also named as
    "*Nucleus Sampling*" as proposed in the paper
    "*The Curious Case of Neural Text Degeneration(Holtzman et al.)*".

    Args:
        start_tokens: 1D :tensor:`LongTensor` shaped ``[batch_size]``,
            representing the start tokens for each sequence in batch.
        end_token: Python int or scalar :tensor:`LongTensor`, denoting the
            token that marks end of decoding.
        p (float, optional): A value used to filter out tokens whose cumulative
            probability is greater than `p` when arranged in decreasing order of
            probabilities. Must be between [0, 1.0]. If set to 1, samples from
            all candidates (i.e., regular random sample decoding). Defaults to
            0.5.
        softmax_temperature (float, optional): Value to divide the logits by
            before computing the softmax. Larger values (above 1.0) result
            in more random samples, while smaller values push the sampling
            distribution towards the argmax. Must be strictly greater than
            0. Defaults to 1.0.

    Raises:
        ValueError: if :attr:`start_tokens` is not a 1D tensor or
            :attr:`end_token` is not a scalar.
    """

    def __init__(self, start_tokens: torch.LongTensor,
                 end_token: Union[int, torch.LongTensor], p: float = 0.9,
                 softmax_temperature: Optional[float] = None):
        super().__init__(start_tokens, end_token)
        self._p = p
        self._softmax_temperature = softmax_temperature

    def sample(self, time: int, outputs: torch.Tensor) -> torch.LongTensor:
        del time  # unused by sample_fn
        # Outputs are logits, we sample from tokens with cumulative
        # probability at most p when arranged in decreasing order
        if not torch.is_tensor(outputs):
            raise TypeError(
                f"Expected outputs to be a single Tensor, got: {type(outputs)}")
        if self._softmax_temperature is None:
            logits = outputs
        else:
            logits = outputs / self._softmax_temperature

        logits = _top_p_logits(logits, p=self._p)

        sample_id_sampler = Categorical(logits=logits)
        sample_ids = sample_id_sampler.sample()

        return sample_ids


class SoftmaxEmbeddingHelper(EmbeddingHelper[torch.Tensor]):
    r"""A helper that feeds softmax probabilities over vocabulary
    to the next step.

    Uses the softmax probability vector to pass through word embeddings to
    get the next input (i.e., a mixed word embedding).

    A subclass of :class:`~texar.torch.modules.Helper`. Used as a helper to
    :class:`~texar.torch.modules.RNNDecoderBase` in inference mode.

    Args:
        embedding: A callable or the ``params`` argument for
            :torch_nn:`functional.embedding`.
            If a callable, it can take a vector tensor of ``ids`` (argmax
            ids), or take two arguments (``ids``, ``times``), where ``ids``
            is a vector of argmax ids, and ``times`` is a vector of current
            time steps (i.e., position ids). The latter case can be used
            when :attr:`embedding` is a combination of word embedding and
            position embedding.
            The returned tensor will be passed to the decoder input.
        start_tokens: 1D :tensor:`LongTensor` shaped ``[batch_size]``,
            representing the start tokens for each sequence in batch.
        end_token: Python int or scalar :tensor:`LongTensor`, denoting the
            token that marks end of decoding.
        tau: A float scalar tensor, the softmax temperature.
        stop_gradient (bool): Whether to stop the gradient backpropagation
            when feeding softmax vector to the next step.
        use_finish (bool): Whether to stop decoding once :attr:`end_token`
            is generated. If `False`, decoding will continue until
            :attr:`max_decoding_length` of the decoder is reached.

    Raises:
        ValueError: if :attr:`start_tokens` is not a 1D tensor or
            :attr:`end_token` is not a scalar.
    """

    def __init__(self, start_tokens: torch.LongTensor,
                 end_token: Union[int, torch.LongTensor], tau: float,
                 stop_gradient: bool = False, use_finish: bool = True):
        super().__init__(start_tokens, end_token)

        self._tau = tau
        self._stop_gradient = stop_gradient
        self._use_finish = use_finish

    def sample(self, time: int, outputs: torch.Tensor) -> torch.Tensor:
        r"""Returns ``sample_id`` which is softmax distributions over vocabulary
        with temperature :attr:`tau`. Shape = ``[batch_size, vocab_size]``.
        """
        del time
        sample_ids = torch.softmax(outputs / self._tau, dim=-1)
        return sample_ids

    def next_inputs(self, embedding_fn: EmbeddingFn,
                    time: int, outputs: torch.Tensor,
                    sample_ids: torch.LongTensor) -> NextInputTuple:
        del outputs  # unused by next_inputs_fn
        if self._use_finish:
            hard_ids = torch.argmax(sample_ids, dim=-1)
            finished = (hard_ids == self._end_token)
        else:
            finished = torch.zeros_like(self._start_tokens, dtype=torch_bool)
        if self._stop_gradient:
            sample_ids = sample_ids.detach()

        indices = torch.arange(sample_ids.size(-1), device=sample_ids.device)
        times = torch.full_like(indices, time + 1)
        embeddings = embedding_fn(indices, times)

        next_inputs = torch.matmul(sample_ids, embeddings)
        return (finished, next_inputs)


class GumbelSoftmaxEmbeddingHelper(SoftmaxEmbeddingHelper):
    r"""A helper that feeds Gumbel softmax sample to the next step.

    Uses the Gumbel softmax vector to pass through word embeddings to
    get the next input (i.e., a mixed word embedding).

    A subclass of :class:`~texar.torch.modules.Helper`. Used as a helper to
    :class:`~texar.torch.modules.RNNDecoderBase` in inference mode.

    Same as :class:`~texar.torch.modules.SoftmaxEmbeddingHelper` except that
    here Gumbel softmax (instead of softmax) is used.

    Args:
        embedding: A callable or the ``params`` argument for
            :torch_nn:`functional.embedding`.
            If a callable, it can take a vector tensor of ``ids`` (argmax
            ids), or take two arguments (``ids``, ``times``), where ``ids``
            is a vector of argmax ids, and ``times`` is a vector of current
            time steps (i.e., position ids). The latter case can be used
            when :attr:`embedding` is a combination of word embedding and
            position embedding.
            The returned tensor will be passed to the decoder input.
        start_tokens: 1D :tensor:`LongTensor` shaped ``[batch_size]``,
            representing the start tokens for each sequence in batch.
        end_token: Python int or scalar :tensor:`LongTensor`, denoting the
            token that marks end of decoding.
        tau: A float scalar tensor, the softmax temperature.
        straight_through (bool): Whether to use straight through gradient
            between time steps. If `True`, a single token with highest
            probability (i.e., greedy sample) is fed to the next step and
            gradient is computed using straight through. If `False`
            (default), the soft Gumbel-softmax distribution is fed to the
            next step.
        stop_gradient (bool): Whether to stop the gradient backpropagation
            when feeding softmax vector to the next step.
        use_finish (bool): Whether to stop decoding once :attr:`end_token`
            is generated. If `False`, decoding will continue until
            :attr:`max_decoding_length` of the decoder is reached.

    Raises:
        ValueError: if :attr:`start_tokens` is not a 1D tensor or
            :attr:`end_token` is not a scalar.
    """

    def __init__(self, start_tokens: torch.LongTensor,
                 end_token: Union[int, torch.LongTensor], tau: float,
                 straight_through: bool = False,
                 stop_gradient: bool = False, use_finish: bool = True):
        super().__init__(start_tokens, end_token, tau,
                         stop_gradient, use_finish)
        self._straight_through = straight_through
        # unit-scale, zero-location Gumbel distribution
        self._gumbel = Gumbel(loc=torch.tensor(0.0), scale=torch.tensor(1.0))

    def sample(self, time: int, outputs: torch.Tensor) -> torch.Tensor:
        r"""Returns ``sample_id`` of shape ``[batch_size, vocab_size]``. If
        :attr:`straight_through` is `False`, this contains the Gumbel softmax
        distributions over vocabulary with temperature :attr:`tau`. If
        :attr:`straight_through` is `True`, this contains one-hot vectors of
        the greedy samples.
        """
        gumbel_samples = self._gumbel.sample(outputs.size()).to(
            device=outputs.device, dtype=outputs.dtype)
        sample_ids = torch.softmax(
            (outputs + gumbel_samples) / self._tau, dim=-1)
        if self._straight_through:
            argmax_ids = torch.argmax(sample_ids, dim=-1).unsqueeze(1)
            sample_ids_hard = torch.zeros_like(sample_ids).scatter_(
                dim=-1, index=argmax_ids, value=1.0)  # one-hot vectors
            sample_ids = (sample_ids_hard - sample_ids).detach() + sample_ids
        return sample_ids


def default_helper_train_hparams():
    r"""Returns default hyperparameters of an RNN decoder helper in the training
    phase.

    See also
    :meth:`~texar.torch.modules.decoders.rnn_decoder_helpers.get_helper` for
    information of the hyperparameters.

    Returns:
        dict: A dictionary with following structure and values:

        .. code-block:: python

            {
                # The `helper_type` argument for `get_helper`, i.e., the name
                # or full path to the helper class.
                "type": "TrainingHelper",

                # The `**kwargs` argument for `get_helper`, i.e., additional
                # keyword arguments for constructing the helper.
                "kwargs": {}
            }
    """
    return {
        'type': 'TrainingHelper',
        'kwargs': {}
    }


def default_helper_infer_hparams():
    r"""Returns default hyperparameters of an RNN decoder helper in the
    inference phase.

    See also
    :meth:`~texar.torch.modules.decoders.rnn_decoder_helpers.get_helper` for
    information of the hyperparameters.

    Returns:
        dict: A dictionary with following structure and values:

        .. code-block:: python

            {
                # The `helper_type` argument for `get_helper`, i.e., the name
                # or full path to the helper class.
                "type": "SampleEmbeddingHelper",

                # The `**kwargs` argument for `get_helper`, i.e., additional
                # keyword arguments for constructing the helper.
                "kwargs": {}
            }
    """
    return {
        'type': 'SampleEmbeddingHelper',
        'kwargs': {}
    }


T = TypeVar('T')  # type argument


def get_helper(helper_type: Union[Type[T], T, str],
               start_tokens: Optional[torch.LongTensor] = None,
               end_token: Optional[Union[int, torch.LongTensor]] = None,
               **kwargs):
    r"""Creates a Helper instance.

    Args:
        helper_type: A :class:`~texar.torch.modules.Helper` class, its
            name or module path, or a class instance. If a class instance
            is given, it is returned directly.
        start_tokens: 1D :tensor:`LongTensor` shaped ``[batch_size]``,
            representing the start tokens for each sequence in batch.
        end_token: Python int or scalar :tensor:`LongTensor`, denoting the
            token that marks end of decoding.
        **kwargs: Additional keyword arguments for constructing the helper.

    Returns:
        A helper instance.
    """
    module_paths = [
        'texar.torch.modules.decoders.decoder_helpers',
        'texar.torch.custom']
    class_kwargs = {'start_tokens': start_tokens,
                    'end_token': end_token}
    class_kwargs.update(kwargs)
    return utils.check_or_get_instance_with_redundant_kwargs(
        helper_type, class_kwargs, module_paths)



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various entropies.
"""

from typing import Optional

import torch
import torch.nn.functional as F

from texar.torch.losses.losses_utils import mask_and_reduce, reduce_dimensions
from texar.torch.utils.shapes import get_rank

__all__ = [
    "entropy_with_logits",
    "sequence_entropy_with_logits",
]


def _get_entropy(logits: torch.Tensor) -> torch.Tensor:
    r"""Compute entropy according to the definition.

    Args:
        logits: Unscaled log probabilities.

    Return:
        A tensor containing the Shannon entropy in the last dimension.
    """
    probs = F.softmax(logits, -1) + 1e-8
    entropy = - probs * torch.log(probs)
    entropy = torch.sum(entropy, -1)
    return entropy


def entropy_with_logits(logits: torch.Tensor,
                        rank: Optional[int] = None,
                        average_across_batch: bool = True,
                        average_across_remaining: bool = False,
                        sum_over_batch: bool = False,
                        sum_over_remaining: bool = True) -> torch.Tensor:
    r"""Shannon entropy given logits.

    Args:
        logits: Unscaled log probabilities of shape
            `[batch_size, d_2, ..., d_{rank-1}, distribution_dim]`
            and of dtype `float32` or `float64`.

            The rank of the tensor is optionally specified by the argument
            :attr:`rank`.

            The tensor is considered as having `[batch_size, .., d_{rank-1}]`
            elements, each of which has a distribution of length `d_rank`
            (i.e., `distribution_dim`). So the last dimension is always
            summed out to compute the entropy.
        rank (int, optional): The rank of :attr:`logits`.
            If `None` (default), `rank` is inferred automatically from
            `logits`. If the inference fails, `rank` is
            set to 2, i.e., assuming :attr:`logits` is of shape
            `[batch_size, distribution_dim]`
        average_across_batch (bool): If set, average the entropy across the
            batch dimension. Must not set `average_across_batch`'
            and `sum_over_batch` at the same time.
        average_across_remaining (bool): If set, average the entropy across the
            remaining dimensions. Must not set `average_across_remaining`'
            and `sum_over_remaining` at the same time.
            Used only when :attr:`logits` has rank >= 3.
        sum_over_batch (bool): If set, sum the entropy across the
            batch dimension. Must not set `average_across_batch`
            and `sum_over_batch` at the same time.
        sum_over_remaining (bool): If set, sum the entropy across the
            remaining dimension. Must not set `average_across_remaining`
            and `sum_over_remaining` at the same time.
            Used only when :attr:`logits` has rank >= 3.

    Returns:
        A Tensor containing the Shannon entropy. The dimensionality of the
        Tensor depends on the configuration of reduction arguments. For
        example, if both batch and remaining dimensions are reduced (by
        either sum or average), the returned Tensor is a scalar Tensor.
    """
    entropy = _get_entropy(logits)

    if rank is None:
        rank = get_rank(logits)
    if rank is None:
        rank = 2
    rank -= 1

    if average_across_batch and sum_over_batch:
        raise ValueError("Only one of `average_across_batch` and "
                         "`sum_over_batch` can be set.")
    if average_across_remaining and sum_over_remaining:
        raise ValueError("Only one of `average_across_remaining` and "
                         "`sum_over_remaining` can be set.")
    sum_axes, average_axes = [], []
    if sum_over_batch:
        sum_axes.append(0)
    if average_across_batch:
        average_axes.append(0)
    if sum_over_remaining and rank >= 2:
        sum_axes += list(range(1, rank))
    if average_across_remaining and rank >= 2:
        average_axes += list(range(1, rank))

    entropy = reduce_dimensions(
        entropy, average_axes=average_axes, sum_axes=sum_axes
    )

    return entropy


def sequence_entropy_with_logits(
        logits: torch.Tensor,
        rank: Optional[int] = None,
        sequence_length: Optional[torch.LongTensor] = None,
        average_across_batch: bool = True,
        average_across_timesteps: bool = False,
        average_across_remaining: bool = False,
        sum_over_batch: bool = False,
        sum_over_timesteps: bool = True,
        sum_over_remaining: bool = True,
        time_major: bool = False) -> torch.Tensor:
    r"""Shannon entropy given logits.

    Args:
        logits: Unscaled log probabilities of shape
            `[batch_size, max_time, d_3, ..., d_{rank-1}, distribution_dim]`
            and of dtype `float32` or `float64`.

            The rank of the tensor is optionally specified by the argument
            :attr:`rank`.

            The tensor is considered as having `[batch_size, .., d_{rank-1}]`
            elements, each of which has a distribution of length `d_rank`
            (i.e., `distribution_dim`). So the last dimension is always
            summed out to compute the entropy.

            The batch and time dimensions are exchanged if :attr:`time_major`
            is `True`.
        rank (int, optional): The rank of :attr:`logits`.
            If `None` (default), `rank` is inferred automatically from
            `logits`. If the inference fails, `rank` is
            set to 3, i.e., assuming `logits` is of shape
            `[batch_size, max_time, distribution_dim]`
        sequence_length (optional): A Tensor of shape `[batch_size]`.
            Time steps beyond the respective sequence lengths are
            counted into the entropy.
        average_across_timesteps (bool): If set, average the entropy across
            the time dimension. Must not set `average_across_timesteps`
            and `sum_over_timesteps` at the same time.
        average_across_batch (bool): If set, average the entropy across the
            batch dimension. Must not set `average_across_batch`'
            and `sum_over_batch` at the same time.
        average_across_remaining (bool): If set, average the entropy across the
            remaining dimensions. Must not set `average_across_remaining`'
            and `sum_over_remaining` at the same time.
            Used only when :attr:`logits` has rank >= 4.
        sum_over_timesteps (bool): If set, sum the entropy across the
            time dimension. Must not set `average_across_timesteps`
            and `sum_over_timesteps` at the same time.
        sum_over_batch (bool): If set, sum the entropy across the
            batch dimension. Must not set `average_across_batch`
            and `sum_over_batch` at the same time.
        sum_over_remaining (bool): If set, sum the entropy across the
            remaining dimension. Must not set `average_across_remaining`
            and `sum_over_remaining` at the same time.
            Used only when :attr:`logits` has rank >= 4.
        time_major (bool): The shape format of the inputs. If `True`,
            :attr:`logits` must have shape `[max_time, batch_size, ...]`.
            If `False` (default), it must have shape
            `[batch_size, max_time, ...]`.

    Returns:
        A Tensor containing the Shannon entropy. The dimensionality of the
        Tensor depends on the configuration of reduction arguments. For
        example, if batch, time, and remaining dimensions are all reduced (by
        either sum or average), the returned Tensor is a scalar Tensor.
    """
    entropy = _get_entropy(logits)

    if rank is None:
        rank = get_rank(logits)
    if rank is None:
        rank = 3
    rank -= 1

    entropy = mask_and_reduce(
        entropy,
        sequence_length,
        rank=rank,
        average_across_batch=average_across_batch,
        average_across_timesteps=average_across_timesteps,
        average_across_remaining=average_across_remaining,
        sum_over_batch=sum_over_batch,
        sum_over_timesteps=sum_over_timesteps,
        sum_over_remaining=sum_over_remaining,
        time_major=time_major
    )

    return entropy
